/*=============================================================================
	ALAudioSubsystem.cpp: Unreal OpenAL Audio interface object.
	Copyright 1999 Epic Games, Inc. All Rights Reserved.

Revision history:
	* Created by Joseph I. Valenzuela (based on AudioSubsystem.cpp)
=============================================================================*/

/*------------------------------------------------------------------------------------
	Audio includes.
------------------------------------------------------------------------------------*/

#include "ALAudio.h"
#include "MikModInterface.h"

#include <AL/al.h>
#include <AL/alc.h>

#include <mikmod.h>
#include <stdio.h>

#define PROCADDRESS(x) (alGetProcAddress((const ALubyte *) x ## "_LOKI"))

#ifndef AL_GAIN_LINEAR
#warning Update your openal
#endif

#define DEFMUSICBUFFERSIZE 16384

/*------------------------------------------------------------------------------------
	UALAudioSubsystem.
------------------------------------------------------------------------------------*/

IMPLEMENT_CLASS(UALAudioSubsystem);

/*
 * PlayingSounds and the associated data type are used in the queue of
 * playing sounds.
 */
typedef struct PlayingSounds_s {
	AActor *Actor;
	USound *Sound;
	FLOAT Priority;
	INT Id;
	ALuint sid;
	FLOAT Volume;
	INT Channel;
} PlayingSound_t;

PlayingSound_t PlayingSounds[MAX_EFFECTS_CHANNELS];

void StopSound( INT Index ) ;
bool sourceIsPlaying(ALuint sid);
static void SetVolumes(UALAudioSubsystem *uas);
static int SmallestPowerOfTwoThatExceedesN(INT N);

UALAudioSubsystem::UALAudioSubsystem()
{
	guard(UALAudioSubsystem::UALAudioSubsystem);

	MusicFade		= 1.0;
	MusicBufferSize         = DEFMUSICBUFFERSIZE;
	CurrentCDTrack		= 255;
	LastTime		= appSeconds();

	IdMapping               = new USoundMap;
	unguard;
}

void UALAudioSubsystem::StaticConstructor()
{
	guard(UALAudioSubsystem::StaticConstructor);

	UEnum* OutputRates = new( GetClass(), TEXT("OutputRates") )UEnum( NULL );
	new( OutputRates->Names )FName( TEXT("8000Hz" ) );
	new( OutputRates->Names )FName( TEXT("11025Hz") );
	new( OutputRates->Names )FName( TEXT("16000Hz") );
	new( OutputRates->Names )FName( TEXT("22050Hz") );
	new( OutputRates->Names )FName( TEXT("32000Hz") );
	new( OutputRates->Names )FName( TEXT("44100Hz") );
	new( OutputRates->Names )FName( TEXT("48000Hz") );
	new(GetClass(),TEXT("UseFilter"),       RF_Public)UBoolProperty  (CPP_PROPERTY(UseFilter      ), TEXT("Audio"), CPF_Config );
	new(GetClass(),TEXT("UseSurround"),     RF_Public)UBoolProperty  (CPP_PROPERTY(UseSurround    ), TEXT("Audio"), CPF_Config );
	new(GetClass(),TEXT("UseStereo"),       RF_Public)UBoolProperty  (CPP_PROPERTY(UseStereo      ), TEXT("Audio"), CPF_Config );
	new(GetClass(),TEXT("UseCDMusic"),      RF_Public)UBoolProperty  (CPP_PROPERTY(UseCDMusic     ), TEXT("Audio"), CPF_Config );
	new(GetClass(),TEXT("UseDigitalMusic"), RF_Public)UBoolProperty  (CPP_PROPERTY(UseDigitalMusic), TEXT("Audio"), CPF_Config );
	new(GetClass(),TEXT("ReverseStereo"),   RF_Public)UBoolProperty  (CPP_PROPERTY(ReverseStereo  ), TEXT("Audio"), CPF_Config );
	new(GetClass(),TEXT("Latency"),         RF_Public)UIntProperty   (CPP_PROPERTY(Latency        ), TEXT("Audio"), CPF_Config );
	new(GetClass(),TEXT("OutputRate"),      RF_Public)UByteProperty  (CPP_PROPERTY(OutputRate     ), TEXT("Audio"), CPF_Config, OutputRates );
	new(GetClass(),TEXT("Channels"), 	RF_Public)UIntProperty   (CPP_PROPERTY(Channels       ), TEXT("Audio"), CPF_Config );
	new(GetClass(),TEXT("MusicVolume"),     RF_Public)UByteProperty  (CPP_PROPERTY(MusicVolume    ), TEXT("Audio"), CPF_Config );
	new(GetClass(),TEXT("SoundVolume"),     RF_Public)UByteProperty  (CPP_PROPERTY(SoundVolume    ), TEXT("Audio"), CPF_Config );
	new(GetClass(),TEXT("AmbientFactor"),   RF_Public)UFloatProperty (CPP_PROPERTY(AmbientFactor  ), TEXT("Audio"), CPF_Config );
	new(GetClass(),TEXT("DopplerSpeed"),    RF_Public)UFloatProperty (CPP_PROPERTY(DopplerSpeed   ), TEXT("Audio"), CPF_Config );
	new(GetClass(),TEXT("MusicBufferSize"), RF_Public)UIntProperty   (CPP_PROPERTY(MusicBufferSize), TEXT("Audio"), CPF_Config );

	unguard;
}

/*------------------------------------------------------------------------------------
	UObject Interface.
------------------------------------------------------------------------------------*/

void UALAudioSubsystem::PostEditChange()
{
	guard(UALAudioSubsystem::PostEditChange);

	// Validate configurable variables.
	OutputRate      = Clamp(OutputRate,(BYTE)0,(BYTE)6);
	Latency         = Clamp(Latency,10,250);
	Channels 	= Clamp(Channels,0,MAX_EFFECTS_CHANNELS);
	DopplerSpeed    = Clamp(DopplerSpeed,1.f,100000.f);
	AmbientFactor   = Clamp(AmbientFactor,0.f,10.f);

	SetVolumes(this);

	unguard;
}

void UALAudioSubsystem::Destroy()
{
	guard(UALAudioSubsystem::Destroy);
	if( Initialized )
	{
		// Unhook.
		USound::Audio = NULL;
		UMusic::Audio = NULL;

		// Shut down viewport.
		SetViewport( NULL );

		// Stop CD.
		if( UseCDMusic && CurrentCDTrack != 255 )
			;//StopCDAudio();

		MikMod_Exit();

		// Shutdown soundsystem.
		if(context_id != NULL) {
			alcDestroyContext(context_id);
			context_id = NULL;
		}

		delete IdMapping;


		debugf( NAME_Exit, TEXT("OpenAL Audio subsystem shut down.") );
	}
	Super::Destroy();
	unguard;
}

void UALAudioSubsystem::ShutdownAfterError()
{
	guard(UALAudioSubsystem::ShutdownAfterError);

	// Unhook.
	USound::Audio = NULL;
	UMusic::Audio = NULL;

	// Safely shut down.
	debugf( NAME_Exit, TEXT("UALAudioSubsystem::ShutdownAfterError") );
	//safecall(AudioStopOutput());
	if( Viewport )
		;//safecall(AudioShutdown());
	Super::ShutdownAfterError();
	unguard;
}

/*------------------------------------------------------------------------------------
	UAudioSubsystem Interface.
------------------------------------------------------------------------------------*/

UBOOL UALAudioSubsystem::Init()
{
	guard(UALAudioSubsystem::Init);
	void (*bombOnError)(void);
	void (*attScale)(ALfloat );

	// Initialize OpenAL Audio library.
	INT Rates[] = {8000, 11025, 16000, 22050, 32000, 44100, 48000};
	INT Rate = Rates[OutputRate];
	INT attrlist[] = { ALC_FREQUENCY, 22050, ALC_BUFFERSIZE, 4096, ALC_INVALID };

	attrlist[1] = Rate;
	attrlist[3] = SmallestPowerOfTwoThatExceedesN(Rate / 6);

	context_id = alcCreateContext(attrlist);
	if(context_id == NULL) {
		return false;
	}

	bombOnError = (void (*)(void))    PROCADDRESS("alBombOnError");
	attScale    = (void (*)(ALfloat)) PROCADDRESS("alAttenuationScale");

#ifdef DEBUG_SOUND
	fprintf(stderr, "BOMB ON ERROR\n");

	// we want to bomb on error
	bombOnError();
#endif

	// set attenuation
	// FIXME: is this right?
	attScale(32000.0 / 60.0);

	// set orientation
	ALfloat zup[] = { 0.0, -1.0, 0.0, 0.0, 0.0, -1.0 };
	alListenerfv(AL_ORIENTATION, zup);

	// set MusicUSound to NULL
	MusicUSound = NULL;

	if( UseDigitalMusic ) {
		char MikModCmdLine[128];

		MikMod_RegisterAllDrivers();
		MikMod_RegisterAllLoaders();

		md_mode |= DMODE_SOFT_MUSIC;
		md_device = MikMod_DriverFromAlias("openal");

		if(md_device == 0) {
			UseDigitalMusic = false;

#ifdef DEBUG_SOUND
			fprintf(stderr, "Could not use Digital Music\n");
#endif
		}

		MusicBufferSize = Clamp(MusicBufferSize, 2048, 65536);
		sprintf(MikModCmdLine,
			"openal-bufsize=%d\n", MusicBufferSize);

		if (MikMod_Init(MikModCmdLine)) {
			UseDigitalMusic = false;

#ifdef DEBUG_SOUND
			fprintf(stderr,
				"Could not initialize sound, reason: %s\n",
				MikMod_strerror(MikMod_errno));
#endif
			return 1;
		}
	}

	// Initialized!
	USound::Audio = this;
	UMusic::Audio = this;
	Initialized = 1;

	debugf( NAME_Init, TEXT("OpenAL Audio subsystem initialized.") );

	SetVolumes(this);

	return true;
	unguard;
}
void UALAudioSubsystem::SetViewport( UViewport* InViewport )
{
	guard(UALAudioSubsystem::SetViewport);

	// Stop playing sounds.
	for( INT i=0; i<Channels; i++ )
		StopSound( i );

	// Remember the viewport.
	if( Viewport != InViewport )
	{
		if( Viewport )
		{
			//unreal_stop_song(0);

			// Unregister everything.
			for( TObjectIterator<UMusic> MusicIt; MusicIt; ++MusicIt )
				if( MusicIt->Handle )
					UnregisterMusic( *MusicIt );

			// Shut down.
			//safecall(AudioStopOutput());
		}
		Viewport = InViewport;
		if( Viewport )
		{
			// Determine startup parameters.
			if( Viewport->Actor->Song && Viewport->Actor->Transition==MTRAN_None )
				Viewport->Actor->Transition = MTRAN_Instant;
		}
	}
	unguard;
}

UViewport* UALAudioSubsystem::GetViewport()
{
	guard(UALAudioSubsystem::GetViewport);
	return Viewport;
	unguard;
}

void UALAudioSubsystem::RegisterSound( USound* Sound )
{
	guard(UALAudioSubsystem::RegisterSound);
	checkSlow(Sound);

	int id;

	if( !Sound->Handle )
	{
		// Set the handle to avoid reentrance.
		Sound->Handle = (void*)-1;

		// Load the data.
		Sound->Data.Load();
		debugf( NAME_DevSound, TEXT("Register sound: %s (%i)"), Sound->GetPathName(), Sound->Data.Num() );
		check(Sound->Data.Num()>0);

		// tsao: add sound.  the push should generate bid
		id = IdMapping->add(Sound);

		if(id == 0) {
			appErrorf( TEXT("Invalid sound format in %s"), Sound->GetFullName() );
		}

		Sound->Handle = (void *) id;

		// Unload the data.
		Sound->Data.Unload();
#ifdef DEBUG_SOUND
		fprintf(stderr,
			"REGISTERed (name %s handle %d):\n",
			Sound->GetPathName(), (int) Sound->Handle);
#endif
	} else {
#ifdef DEBUG_SOUND
		fprintf(stderr,
			"skip REGISTERed (name %s handle %d):\n",
			Sound->GetPathName(), (int) Sound->Handle);
#endif
	}
	unguard;
}

void UALAudioSubsystem::UnregisterSound( USound* Sound )
{
	guard(UALAudioSubsystem::UnregisterSound);
	check(Sound);
	if( Sound->Handle )
	{
		debugf( NAME_DevSound, TEXT("Unregister sound: %s"), Sound->GetFullName() );

		// Stop this sound.
		for( INT i=0; i<Channels; i++ )
			if ( PlayingSounds[i].Sound == Sound )
				StopSound( i );

		// Unload this sound.
		//safecall( UnloadSample( (Sample*) Sound->Handle ) );

		IdMapping->remove(Sound);
	}
	unguard;
}

void UALAudioSubsystem::UnregisterMusic( UMusic* Music )
{
	guard(UALAudioSubsystem::UnregisterMusic);

	stopMusic(Music);
	MusicUSound = NULL;

	unguard;
}

UBOOL UALAudioSubsystem::Exec( const TCHAR* Cmd, FOutputDevice& Ar )
{
	guard(UALAudioSubsystem::Exec);

	const TCHAR* Str = Cmd;
	if( ParseCommand(&Str,TEXT("ASTAT")) )
	{
		if( ParseCommand(&Str,TEXT("Audio")) )
		{
			AudioStats ^= 1;
			return 1;
		}
		if( ParseCommand(&Str,TEXT("Detail")) )
		{
			DetailStats ^= 1;
			return 1;
		}
	}
	return 0;
	
	unguard;
}

UBOOL UALAudioSubsystem::PlaySound
(
	AActor*	Actor,
	INT	Id,
	USound*	Sound,
	FVector	Location,
	FLOAT	Volume,
	FLOAT	Radius,
	FLOAT	Pitch
)
{
	guard(UALAudioSubsystem::PlaySound);
	check(Radius);
	if( !Viewport || !Sound ) {
		return 0;
	}

#ifdef DEBUG_SOUND
	fprintf(stderr, "Play %s\n", Sound->GetPathName());
#endif

	// Allocate a new slot if requested.
	if( (Id&14)==2*SLOT_None )
		Id = 16 * --FreeSlot;

	// Compute priority.
	FLOAT Priority = SoundPriority( Viewport, Location, Volume, Radius );

	// If already playing, stop it.
	INT   Index        = -1;
	FLOAT BestPriority = Priority;
	for( INT i=0; i<Channels; i++ )
	{
		PlayingSound_t Playing = PlayingSounds[i];

		if( (Playing.Id&~1)==(Id&~1) )
		{
			// Skip if not interruptable.
			if( Id&1 ) {
				return 0;
			}

			// Stop the sound.
			Index = i;
			break;
		}
		else if( Playing.Priority<=BestPriority )
		{
			Index = i;
			BestPriority = Playing.Priority;
		}
	}

	// If no sound, or its priority is overruled, stop it.
	if( Index==-1 ) {

		return 0;
	}


	// Put the sound on the play-list.
	StopSound( Index );
	if( Sound!=(USound*)-1 ) {
		ALuint sid;
		ALuint bid;

		bid = IdMapping->find(Sound);
		if(bid == 0) {
			bid = IdMapping->add(Sound);

			if(bid == 0) {

				return 0;
			}

			Sound->Handle = (void *) bid;
		}


		alGenSources(1, &sid);

		alSourcei(sid, AL_BUFFER, bid);

		/* set volume */
		ALfloat fvol = Volume / 2.0f;

		fvol *= ((float) SoundVolume / AUDIO_MAXVOLUME);

		alSourcef(sid, AL_GAIN_LINEAR, fvol);

		alSourcePlay(sid);

		if(Actor != NULL) {
			FVector Location = Actor->Location;

			alSource3f(sid, AL_POSITION,
				Location.X, Location.Y, Location.Z);
		}

		PlayingSounds[Index].Sound    = Sound;
		PlayingSounds[Index].Actor    = Actor;
		PlayingSounds[Index].Priority = Priority;
		PlayingSounds[Index].Id       = Id;
		PlayingSounds[Index].sid      = sid;
	}

	return 1;
	unguard;
}

void UALAudioSubsystem::NoteDestroy( AActor* Actor )
{
	guard(UALAudioSubsystem::NoteDestroy);
	check(Actor);
	check(Actor->IsValid());

	// Stop referencing actor.
	for( INT i=0; i<Channels; i++ )
	{
		if( PlayingSounds[i].Actor==Actor )
		{
			if( (PlayingSounds[i].Id&14)==SLOT_Ambient*2 )
			{
				// Stop ambient sound when actor dies.
				StopSound( i );
			}
			else
			{
				// Unbind regular sounds from actors.
				PlayingSounds[i].Actor = NULL; 



			}
		}
	}

	unguard;
}

void UALAudioSubsystem::RenderAudioGeometry( FSceneNode* Frame )
{
	guard(UALAudioSubsystem::RenderAudioGeometry);

	unguard;
}

void UALAudioSubsystem::Update( FPointRegion Region, FCoords& Coords )
{
	guard(UALAudioSubsystem::Update);

	if( !Viewport )
		return;


	ALfloat lispos[] = { 0., 0., 0. };

	lispos[0] = Viewport->Actor->Location.X;
	lispos[1] = Viewport->Actor->Location.Y;
	lispos[2] = Viewport->Actor->Location.Z;

	alListenerfv(AL_POSITION, lispos);

	// Nicholas Vinen - now is the time to change music.
	// I didn't implement transitions. Does anybody miss them?

	// vogel: I guess most of this code is crap...
	if ( Viewport->Actor->Transition != MTRAN_None && UseDigitalMusic )
	{
		if (Viewport->Actor->Song)
		{
			int transitiontime;

			/* Work out what music to play and start playing it */
			Viewport->Actor->Song->Data.Load();
			switch(Viewport->Actor->Transition)
			{
			case MTRAN_Fade:
				transitiontime = 1000;
				break;
			case MTRAN_SlowFade:
				transitiontime = 5000;
				break;
			case MTRAN_FastFade:
				transitiontime = 333;
				break;
			default:
				transitiontime = 0;
				break;
			}

			/* stop music if playing */
			if(MusicUSound != NULL) {
				stopMusic(MusicUSound);
			}

			MusicUSound = Viewport->Actor->Song;

			int err = playMusic(MusicUSound);
			if(err == 0) {
				MusicUSound = NULL;
			}
		} else {
			stopMusic(MusicUSound);

			MusicUSound = NULL;
		}

		Viewport->Actor->Transition = MTRAN_None;
	}

	/* music, get correct volume */
	md_musicvolume = (MusicVolume/2);

	if(Player_Active() == 0) {
		/* Music must be over */
		MikMod_Update();

		if(MusicUSound != NULL) {
			/* loop the music */
			int err = playMusic(MusicUSound);
			if(err == 0) {
				MusicUSound = NULL;
			}
		}
	} else {
		if(MusicUSound != NULL) {
			MikMod_Update();
		}
	}

	for(INT i = 0 ; i < Channels; i++ ) {
		PlayingSound_t &Playing = PlayingSounds[i];
		ALuint sid              = Playing.sid;

		if(alIsSource(sid) == AL_TRUE) {
			if(sourceIsPlaying(sid) == false) {
				alDeleteSources(1, &sid);

				Playing.Sound = NULL;
				Playing.Actor = NULL;
				Playing.sid   = 0;
				Playing.Id    = 0;
				Playing.Priority = 0.0F;
			} else {
				/* Update position, gain etc */
				if( Playing.Actor == NULL) {
					continue;
				}

				// Update positioning from actor, if available.
				FVector Location = Playing.Actor->Location;
				alSource3f(sid, AL_POSITION,
					Location.X, Location.Y, Location.Z);
			}
		}
	}

	AActor *ViewActor = Viewport->Actor->ViewTarget?Viewport->Actor->ViewTarget:Viewport->Actor;

	// See if any new ambient sounds need to be started.
	UBOOL Realtime = Viewport->IsRealtime() && Viewport->Actor->Level->Pauser==TEXT("");
	if( Realtime )
	{
		guard(StartAmbience);
		for( INT i=0; i<Viewport->Actor->GetLevel()->Actors.Num(); i++ )
		{
			AActor* Actor = Viewport->Actor->GetLevel()->Actors(i);
			if
			(	Actor
			&&	Actor->AmbientSound
			&&	FDistSquared(ViewActor->Location,Actor->Location)<=Square(Actor->WorldSoundRadius()) )
			{
				INT Id = Actor->GetIndex()*16+SLOT_Ambient*2;
				for( INT j=0; j<Channels; j++ )
					if( PlayingSounds[j].Id==Id )
						break;
				if( j==Channels )
					PlaySound( Actor, Id, Actor->AmbientSound, Actor->Location, AmbientFactor*Actor->SoundVolume/255.0, Actor->WorldSoundRadius(), Actor->SoundPitch/64.0 );
			}
		}
		unguard;
	}

	// tsao: STUB STUB STUB
	unguard;
}

void UALAudioSubsystem::PostRender( FSceneNode* Frame )
{
	guard(UALAudioSubsystem::PostRender);

	// STUB STUB STUB

	unguard;
}

/*------------------------------------------------------------------------------------
	Internals.
------------------------------------------------------------------------------------*/

void StopSound( INT Index )
{
	guard(UALAudioSubsystem::StopSound);

	ALuint sid;

	sid = PlayingSounds[Index].sid;

	if(alIsSource(sid) == AL_TRUE) {
		if(sourceIsPlaying(sid) == true) {
			alSourceStop(sid);
		}

		alDeleteSources(1, &sid);
	}


	PlayingSounds[Index].Sound    = NULL;
	PlayingSounds[Index].Actor    = NULL;
	PlayingSounds[Index].Id       = 0;
	PlayingSounds[Index].sid      = 0;
	PlayingSounds[Index].Priority = 0.0F;

	return;

	unguard;
}

bool sourceIsPlaying(ALuint sid) {
	ALint state;

	alGetSourcei(sid, AL_SOURCE_STATE, &state);

	if(state == AL_PLAYING) {
		return true;
	}

	return false;
}

static void SetVolumes(UALAudioSubsystem *uas)
{
	alListenerf(AL_GAIN_LINEAR,
		    static_cast<ALfloat>(uas->SoundVolume) / AUDIO_MAXVOLUME);

	md_musicvolume = (uas->MusicVolume/2);

	MikMod_Update(); /* allow music changes to take effect */

	return;
}

static int SmallestPowerOfTwoThatExceedesN(INT N)
{
	int retval = 1;

	while(retval < N) {
		retval <<= 1;
	}

	return retval;
}
/*-----------------------------------------------------------------------------
	The End.
-----------------------------------------------------------------------------*/
