/*=============================================================================
	SDLDrv.h: Unreal SDL viewport and platform driver.
	Copyright 1999 Epic Games, Inc. All Rights Reserved.

Revision history:
	* Created by Daniel Vogel (based on XDrv).
	* Crippled by Daniel Vogel

=============================================================================*/

#ifndef _INC_SDLDRV
#define _INC_SDLDRV

/*----------------------------------------------------------------------------
	Dependencies.
----------------------------------------------------------------------------*/

// System includes.
#include <stdlib.h>

// Unreal includes.
#include "Engine.h"
#include "UnRender.h"

// SDL includes
#include "SDL.h"


/*-----------------------------------------------------------------------------
	Declarations.
-----------------------------------------------------------------------------*/

// Classes.
class USDLViewport;
class USDLClient;

/*-----------------------------------------------------------------------------
	USDLClient.
-----------------------------------------------------------------------------*/

//
// SDL implementation of the client.
//
class USDLClient : public UClient, public FNotifyHook
{
	DECLARE_CLASS(USDLClient,UClient,CLASS_Transient|CLASS_Config,SDLDrv)

	// Configuration.
	BITFIELD			StartupFullscreen;
	BITFIELD			SlowVideoBuffering;
	BITFIELD			UseJoystick;
	BITFIELD			DeadZoneX;
	BITFIELD			DeadZoneY;
	BITFIELD			DeadZoneZ;
	BITFIELD			DeadZoneR;
	BITFIELD			DeadZoneU;
	BITFIELD			DeadZoneV;
	BITFIELD			InvertVertical;
	BITFIELD			CaptureMouse;
	INT				JoystickNumber;
	FLOAT				ScaleX;
	FLOAT				ScaleY;
	FLOAT				ScaleZ;
	FLOAT				ScaleR;
	FLOAT				ScaleU;
	FLOAT				ScaleV;
	FLOAT				ScaleJBX;
	FLOAT				ScaleJBY;

	// Variables.
	UBOOL				InMenuLoop;
	UBOOL				ConfigReturnFullscreen;
	INT				NormalMouseInfo[3];
	INT				CaptureMouseInfo[3];

	// Joystick.
	SDL_Joystick*			Joystick;
	INT				NumButtons;
	INT				NumAxes;

	// Constructors.
	USDLClient();
	void StaticConstructor();

	// FNotifyHook interface.
	void NotifyDestroy( void* Src );

	// UObject interface.
	void Destroy();
	void PostEditChange();
	void ShutdownAfterError();

	// UClient interface.
	void Init( UEngine* InEngine );
	void ShowViewportWindows( DWORD ShowFlags, INT DoShow );
	void EnableViewportWindows( DWORD ShowFlags, INT DoEnable );
	UBOOL Exec( const TCHAR* Cmd, FOutputDevice& Ar=*GLog );
	void Tick();
	void MakeCurrent( UViewport* InViewport );
	class UViewport* NewViewport( const FName Name );
};

/*-----------------------------------------------------------------------------
	USDLViewport.
-----------------------------------------------------------------------------*/

//
// Viewport window status.
//
enum ESDLViewportStatus
{
	SDL_ViewportOpening	= 0, // Viewport is opening.
	SDL_ViewportNormal	= 1, // Viewport is operating normally.
	SDL_ViewportClosing	= 2, // Viewport is closing and CloseViewport has been called.
};

//
// A SDL viewport.
//
class USDLViewport : public UViewport
{
	DECLARE_CLASS(USDLViewport,UViewport,CLASS_Transient,SDLDrv)
	DECLARE_WITHIN(USDLClient)

	// Variables.
	ESDLViewportStatus		ViewportStatus;
	INT				HoldCount;
	DWORD				BlitFlags;
	UBOOL				Borderless;
	UBOOL				Mapped;
	UBOOL				Iconified;
	UBOOL				LastKey;
	FTime				RepeatTimer;

	// Info saved during captures and fullscreen sessions.
	INT				SavedColorBytes;
	INT				SavedCaps;

	// SDL Keysym to EInputKey map.
	BYTE				KeysymMap[512];

	// KeyRepeatKey.
        INT				KeyRepeatKey;

	// JoystickButton map
	BYTE				JoystickButtonMap[512];

	// Joystick Hat Hack
	EInputKey 			LastJoyHat;
	
	// Constructor.
	USDLViewport();

	// UObject interface.
	void Destroy();
	void ShutdownAfterError();

	// UViewport interface.
	UBOOL Lock( FPlane FlashScale, FPlane FlashFog, FPlane ScreenClear, DWORD RenderLockFlags, BYTE* HitData=NULL, INT* HitSize=0 );
	UBOOL Exec( const TCHAR* Cmd, FOutputDevice& Ar );
	UBOOL ResizeViewport( DWORD BlitFlags, INT NewX=INDEX_NONE, INT NewY=INDEX_NONE, INT NewColorBytes=INDEX_NONE );
	UBOOL IsFullscreen();
	void Unlock( UBOOL Blit );
	void Repaint( UBOOL Blit );
	void SetModeCursor();
	void UpdateWindowFrame();
	void OpenWindow( DWORD ParentWindow, UBOOL Temporary, INT NewX, INT NewY, INT OpenX, INT OpenY );
	void CloseWindow();
	void UpdateInput( UBOOL Reset );
	void* GetWindow();
	void* GetServer();
	void SetMouseCapture( UBOOL Capture, UBOOL Clip, UBOOL FocusOnly );

	// USDLViewport interface.
	void ToggleFullscreen();
	void EndFullscreen();
	void FindAvailableModes();
	void TryHardware3D( USDLViewport* Viewport );
	UBOOL CauseInputEvent( INT iKey, EInputAction Action, FLOAT Delta=0.0 );
	void TryRenderDevice( const TCHAR* ClassName, INT NewX, INT NewY, INT NewColorBytes, UBOOL Fullscreen );
	void Tick();
	void Iconify();
	// Cursor GetNullCursor();
	void CaptureInputs();
	void ReleaseInputs();

	// Joystick handling.
	UBOOL JoystickInputEvent( FLOAT Delta, EInputKey Key, FLOAT Scale, UBOOL DeadZone );

};

#endif //_INC_SDLDRV
/*-----------------------------------------------------------------------------
	The End.
-----------------------------------------------------------------------------*/
