/*=============================================================================
	SDLViewport.cpp: USDLViewport code.
	Copyright 1999 Epic Games, Inc. All Rights Reserved.

	Revision history:
		* Created by Daniel Vogel (based on XDrv)
		* Changes by John Fulmer		
		* Major changes by Daniel Vogel
	
=============================================================================*/

#include "SDLDrv.h"

/*-----------------------------------------------------------------------------
	Class implementation.
-----------------------------------------------------------------------------*/

IMPLEMENT_CLASS(USDLViewport);

/*-----------------------------------------------------------------------------
	USDLViewport Init/Exit.
-----------------------------------------------------------------------------*/

//
// Constructor.
//
USDLViewport::USDLViewport()
:	UViewport()
,	ViewportStatus( SDL_ViewportOpening )
{
	guard(USDLViewport::USDLViewport);

	// Query current bit depth
	int depth = SDL_GetVideoInfo()->vfmt->BitsPerPixel;
	switch (depth)
	{
		case 8 :	
			ColorBytes = 2;
			break;

		case 15 :	
			ColorBytes = 2;		
			break;

		case 16 :
			ColorBytes = 2;
			Caps |= CC_RGB565;
			break;

		case 24 :
		case 32 :
			ColorBytes = 4;
    			break;

		default :	
			ColorBytes = 2;
			Caps |= CC_RGB565;
	}
	
	// Zero out maps.
	for (INT i=0; i<512; i++)
	{
		KeysymMap[i] 		= 0;
		JoystickButtonMap[i] 	= 0;
	}
		
	// Remap important keys.

	// TTY Functions.
	KeysymMap[SDLK_BACKSPACE]= IK_Backspace;
	KeysymMap[SDLK_TAB]	= IK_Tab;
	KeysymMap[SDLK_RETURN]	= IK_Enter;
	KeysymMap[SDLK_PAUSE]	= IK_Pause;
	KeysymMap[SDLK_ESCAPE]	= IK_Escape;
	KeysymMap[SDLK_DELETE]	= IK_Delete;
	KeysymMap[SDLK_INSERT]  = IK_Insert;

	// Modifiers.
	KeysymMap[SDLK_LSHIFT]	= IK_LShift;
	KeysymMap[SDLK_RSHIFT]	= IK_RShift;
	KeysymMap[SDLK_LCTRL]	= IK_LControl;
	KeysymMap[SDLK_RCTRL]	= IK_RControl;
	KeysymMap[SDLK_LMETA]	= IK_Alt;
	KeysymMap[SDLK_RMETA]	= IK_Alt;
	KeysymMap[SDLK_LALT]	= IK_Alt;
	KeysymMap[SDLK_RALT]	= IK_Alt;
	
	// Special remaps.
	KeysymMap[SDLK_BACKQUOTE]= IK_Tilde;
	KeysymMap[SDLK_LEFTBRACKET]  = IK_LeftBracket;
	KeysymMap[SDLK_RIGHTBRACKET] = IK_RightBracket;
 
	// Misc function keys.
	KeysymMap[SDLK_F1]	= IK_F1;
	KeysymMap[SDLK_F2]	= IK_F2;
	KeysymMap[SDLK_F3]	= IK_F3;
	KeysymMap[SDLK_F4]	= IK_F4;
	KeysymMap[SDLK_F5]	= IK_F5;
	KeysymMap[SDLK_F6]	= IK_F6;
	KeysymMap[SDLK_F7]	= IK_F7;
	KeysymMap[SDLK_F8]	= IK_F8;
	KeysymMap[SDLK_F9]	= IK_F9;
	KeysymMap[SDLK_F10]	= IK_F10;
	KeysymMap[SDLK_F11]	= IK_F11;
	KeysymMap[SDLK_F12]	= IK_F12;
	KeysymMap[SDLK_F13]	= IK_F13;
	KeysymMap[SDLK_F14]	= IK_F14;
	KeysymMap[SDLK_F15]	= IK_F15;

	// Cursor control and motion.
	KeysymMap[SDLK_HOME]	= IK_Home;
	KeysymMap[SDLK_LEFT]	= IK_Left;
	KeysymMap[SDLK_UP]	= IK_Up;
	KeysymMap[SDLK_RIGHT]	= IK_Right;
	KeysymMap[SDLK_DOWN]	= IK_Down;
	KeysymMap[SDLK_PAGEUP]	= IK_PageUp;
	KeysymMap[SDLK_PAGEDOWN]= IK_PageDown;
	KeysymMap[SDLK_END]	= IK_End;

	// Keypad functions and numbers.
	KeysymMap[SDLK_KP_ENTER]= IK_Enter;
	KeysymMap[SDLK_KP0]	= IK_NumPad0;
	KeysymMap[SDLK_KP1]	= IK_NumPad1;
	KeysymMap[SDLK_KP2]	= IK_NumPad2;
	KeysymMap[SDLK_KP3]	= IK_NumPad3;
	KeysymMap[SDLK_KP4]	= IK_NumPad4;
	KeysymMap[SDLK_KP5]	= IK_NumPad5;
	KeysymMap[SDLK_KP6]	= IK_NumPad6;
	KeysymMap[SDLK_KP7]	= IK_NumPad7;
	KeysymMap[SDLK_KP8]	= IK_NumPad8;
	KeysymMap[SDLK_KP9]	= IK_NumPad9;
	KeysymMap[SDLK_KP_MULTIPLY]= IK_GreyStar;
	KeysymMap[SDLK_KP_PLUS]	= IK_GreyPlus;
	// vogel: TODO - what is KP_Separator in SDL???
	// KeysymMap[XK_KP_Separator]	= IK_Separator;
	KeysymMap[SDLK_KP_MINUS]= IK_GreyMinus;
	KeysymMap[SDLK_KP_PERIOD]= IK_NumPadPeriod;
	KeysymMap[SDLK_KP_DIVIDE]= IK_GreySlash;

	// Other
	KeysymMap[SDLK_MINUS]	= IK_Minus;
	KeysymMap[SDLK_EQUALS]	= IK_Equals;     

	KeyRepeatKey = 0;
	LastJoyHat = IK_None;
	debugf( TEXT("Created and initialized a new SDL viewport.") );

	unguard;
}

//
// Destroy.
//
void USDLViewport::Destroy()
{
	guard(USDLViewport::Destroy);
	Super::Destroy();
	unguard;
}

//
// Error shutdown.
//
void USDLViewport::ShutdownAfterError()
{
	debugf( NAME_Exit, TEXT("Executing USDLViewport::ShutdownAfterError") );
	Super::ShutdownAfterError();
}

/*-----------------------------------------------------------------------------
	Command line.
-----------------------------------------------------------------------------*/

//
// Command line.
//
UBOOL USDLViewport::Exec( const TCHAR* Cmd, FOutputDevice& Ar )
{
	guard(USDLViewport::Exec);
	if( UViewport::Exec( Cmd, Ar ) )
	{
		return 1;
	}
	else if( ParseCommand(&Cmd,TEXT("EndFullscreen")) )
	{
		EndFullscreen();
		return 1;
	}
	else if( ParseCommand(&Cmd,TEXT("ToggleFullscreen")) )
	{
		ToggleFullscreen();
		return 1;
	}
	else if( ParseCommand(&Cmd,TEXT("Iconify")) )
	{
		Iconify();
		return 1;
	}
	else if( ParseCommand(&Cmd,TEXT("GetCurrentRes")) )
	{
		Ar.Logf( TEXT("%ix%i"), SizeX, SizeY, (ColorBytes?ColorBytes:2)*8 );
		return 1;
	}
	else if( ParseCommand(&Cmd,TEXT("GetCurrentColorDepth")) )
	{
		Ar.Logf( TEXT("%i"), (ColorBytes?ColorBytes:2)*8 );
		return 1;
	}
	else if( ParseCommand(&Cmd,TEXT("GetColorDepths")) )
	{
		Ar.Logf( TEXT("%i"), ColorBytes * 8 );	// Lock bpp to current
		return 1;
	}
	else if( ParseCommand(&Cmd,TEXT("GetCurrentRenderDevice")) )
	{
		Ar.Log( RenDev->GetClass()->GetPathName() );
		return 1;
	}
	else if( ParseCommand(&Cmd,TEXT("SetRenderDevice")) )
	{
		FString Saved = RenDev->GetClass()->GetPathName();
		INT SavedSizeX=SizeX, SavedSizeY=SizeY, SavedColorBytes=ColorBytes, SavedFullscreen=((BlitFlags & BLIT_Fullscreen)!=0);
		TryRenderDevice( Cmd, SizeX, SizeY, ColorBytes, SavedFullscreen );
		if( !RenDev )
		{
			TryRenderDevice( *Saved, SavedSizeX, SavedSizeY, SavedColorBytes, SavedFullscreen );
			check(RenDev);
			Ar.Log(TEXT("0"));
		}
		else Ar.Log(TEXT("1"));
		return 1;
	}
	else if( ParseCommand(&Cmd,TEXT("GetRes")) )
	{
		return 1;
	}
	else if( ParseCommand(&Cmd,TEXT("SetRes")) )
	{
		INT X=appAtoi(Cmd);
		TCHAR* CmdTemp = (TCHAR*) (appStrchr(Cmd,'x') ? appStrchr(Cmd,'x')+1 : appStrchr(Cmd,'X') ? appStrchr(Cmd,'X')+1 : TEXT(""));
		INT Y=appAtoi(CmdTemp);
		Cmd = CmdTemp;
		CmdTemp = (TCHAR*) (appStrchr(Cmd,'x') ? appStrchr(Cmd,'x')+1 : appStrchr(Cmd,'X') ? appStrchr(Cmd,'X')+1 : TEXT(""));
		INT C=appAtoi(CmdTemp);
		INT NewColorBytes = C ? C/8 : ColorBytes;
		if( X && Y )
		{
			HoldCount++;
			UBOOL Result = RenDev->SetRes( X, Y, NewColorBytes, IsFullscreen() );
			HoldCount--;
			if( !Result )
				EndFullscreen();
		}
		return 1;
	}
	else if( ParseCommand(&Cmd,TEXT("Preferences")) )
	{
		// No preferences window.
	
		return 1;
	}
	else return 0;
	unguard;
}

/*-----------------------------------------------------------------------------
	Window openining and closing.
-----------------------------------------------------------------------------*/

//
// Open this viewport's window.
//
void USDLViewport::OpenWindow( DWORD InParentWindow, UBOOL IsTemporary, INT NewX, INT NewY, INT OpenX, INT OpenY )
{
	guard(USDLViewport::OpenWindow);
	check(Actor);
	check(!HoldCount);
	USDLClient* C = GetOuterUSDLClient();

	debugf( TEXT("Opening SDL viewport.") );

	// Create or update the window.
	SizeX = C->FullscreenViewportX;
	SizeY = C->FullscreenViewportY;

	FindAvailableModes();

	// Create rendering device.
	if( !RenDev && !GIsEditor && !ParseParam(appCmdLine(),TEXT("nohard")) )
		TryRenderDevice( TEXT("ini:Engine.Engine.GameRenderDevice"), NewX, NewY, ColorBytes, C->StartupFullscreen );
	check(RenDev);
	UpdateWindowFrame();
	Repaint( 1 );
	unguard;
}

//
// Close a viewport window.  Assumes that the viewport has been openened with
// OpenViewportWindow.  Does not affect the viewport's object, only the
// platform-specific information associated with it.
//
void USDLViewport::CloseWindow()
{
	guard(USDLViewport::CloseWindow);
	unguard;
}

/*-----------------------------------------------------------------------------
	USDLViewport operations.
-----------------------------------------------------------------------------*/

//
// Find all available DirectDraw modes for a certain number of color bytes.
//
void USDLViewport::FindAvailableModes()
{
	guard(UWindowsViewport::FindAvailableModes);
	unguard;
}

//
// Repaint the viewport.
//
void USDLViewport::Repaint( UBOOL Blit )
{
	guard(USDLViewport::Repaint);
	GetOuterUSDLClient()->Engine->Draw( this, Blit );
	unguard;
}

//
// Return whether fullscreen.
//
UBOOL USDLViewport::IsFullscreen()
{
	guard(USDLViewport::IsFullscreen);
	return (BlitFlags & BLIT_Fullscreen)!=0;
	unguard;
}

//
// Set the mouse cursor according to Unreal or UnrealEd's mode, or to
// an hourglass if a slow task is active.
//
void USDLViewport::SetModeCursor()
{
	guard(USDLViewport::SetModeCursor);
	// FIXME: Implement
	unguard;
}

//
// Update user viewport interface.
//
void USDLViewport::UpdateWindowFrame()
{
	guard(USDLViewport::UpdateWindowFrame);

	// If not a window, exit.
	if( HoldCount || (BlitFlags&BLIT_Fullscreen) || (BlitFlags&BLIT_Temporary) )
		return;

	unguard;
}

//
// Return the viewport's window.
//
void* USDLViewport::GetWindow()
{
	guard(USDLViewport::GetWindow);
     	return NULL;
	unguard;
}

//
// Return the viewport's display.
//
void* USDLViewport::GetServer()
{
	guard(USDLViewport::GetServer);
	return NULL;
	unguard;
}

/*-----------------------------------------------------------------------------
	Input.
-----------------------------------------------------------------------------*/

//
// Input event router.
//
UBOOL USDLViewport::CauseInputEvent( INT iKey, EInputAction Action, FLOAT Delta )
{
	guard(USDLViewport::CauseInputEvent);

	// Route to engine if a valid key; some keyboards produce key
	// codes that go beyond IK_MAX.
	if( iKey>=0 && iKey<IK_MAX )
		return GetOuterUSDLClient()->Engine->InputEvent( this, (EInputKey)iKey, Action, Delta );
	else
		return 0;
	unguard;
}

//
// If the cursor is currently being captured, stop capturing, clipping, and 
// hiding it, and move its position back to where it was when it was initially
// captured.
//
void USDLViewport::SetMouseCapture( UBOOL Capture, UBOOL Clip, UBOOL OnlyFocus )
{
	guard(USDLViewport::SetMouseCapture);

	bWindowsMouseAvailable = 0;//!Capture;

	if( OnlyFocus )
		return;

	if( Capture )
	{
		CaptureInputs();
	} else {
		ReleaseInputs();
	}
	unguard;
}


UBOOL USDLViewport::JoystickInputEvent( FLOAT Delta, EInputKey Key, FLOAT Scale, UBOOL DeadZone )
{
	guard(USDLViewport::JoystickInputEvent);
	Delta /= 32768.0;
	if( DeadZone )
	{
		if( Delta > 0.2 )
			Delta = (Delta - 0.2) / 0.8;
		else if( Delta < -0.2 )
			Delta = (Delta + 0.2) / 0.8;
		else
			Delta = 0.0;
	}
	return CauseInputEvent( Key, IST_Axis, Scale * Delta );
	unguard;
}

//
// Update input for this viewport.
//
void USDLViewport::UpdateInput( UBOOL Reset )
{
	guard(USDLViewport::UpdateInput);

	USDLClient* Client = GetOuterUSDLClient();

	// Keyboard.
	EInputKey Key 	    = IK_None;
	EInputKey UpperCase = IK_None;
	EInputKey LowerCase = IK_None;

	// Mouse movement management.
	UBOOL MouseMoved = false;
	UBOOL JoyHatMoved = false;
	INT DX = 0, DY = 0;
	EInputKey ThisJoyHat = IK_None;

	char* sdl_error;
	SDL_Event Event;
	while( SDL_PollEvent( &Event ) ) 
	{
		switch( Event.type )
		{
		case SDL_KEYDOWN:
			// vogel: this is a MESS - clean this part up!!!
			if ( (Event.key.keysym.sym == SDLK_g) && (Event.key.keysym.mod & KMOD_CTRL) )
			{
				SDL_GrabMode mode = SDL_WM_GrabInput(SDL_GRAB_QUERY);
				if ( mode == SDL_GRAB_ON ) 
				{
					SDL_WM_GrabInput(SDL_GRAB_OFF);
				} else {
					SDL_WM_GrabInput(SDL_GRAB_ON);
				}
				// Don't pass event to engine.
				break;
			}
			// Reset timer.
			RepeatTimer = appSeconds();
			LastKey = true;

			if ( Event.key.keysym.unicode ) {
				Key = (EInputKey) Event.key.keysym.unicode;
			} else {
				Key = (EInputKey) Event.key.keysym.sym;
			}
			
			// Convert to UpperCase/LowerCase values. If not
			// a letter, leave alone
			if ((Key >= ((EInputKey) SDLK_a)) && (Key <= ((EInputKey) SDLK_z))) 
			{
				LowerCase = Key;
				UpperCase = (EInputKey) (Key - 32);
			} 
			else if ((Key >= ((EInputKey) (SDLK_a - 32))) && (Key <= ((EInputKey) (SDLK_z - 32)))) 
			{
				UpperCase = Key;
				LowerCase = (EInputKey) (Key + 32);
			} 
			else 	
			{
				UpperCase = Key;
				LowerCase = Key;
			}
			// Always send upper case letters to the system.
			// The engine NEVER processes lowercase letters, 
			// since lowercase keycodes are mapped to the F keys

			Key = UpperCase;

			// Check the Keysym map.
			if (KeysymMap[Key] != 0)
				Key = (EInputKey) KeysymMap[Key];
	
			// Reset to origional values. In case below gets 
			// munged by the above code. Shouldn't happen.
			if (Key == (EInputKey) SDLK_BACKSPACE)
				Key = IK_Backspace;
			if (Key ==(EInputKey) SDLK_TAB)
				Key = IK_Tab;
			if (Key == (EInputKey) SDLK_RETURN)
				Key = IK_Enter;
			if (Key == (EInputKey) SDLK_DELETE)
				Key = IK_Delete;

			// Send key to input system. 
			if ( Key != IK_None )
				CauseInputEvent( Key, IST_Press );

			if ( Event.key.keysym.mod & KMOD_SHIFT )
				Key = UpperCase;
			else	
				Key = LowerCase;
		
			KeyRepeatKey = Key;
		
			// Send to text processor	
			if (	(Event.key.keysym.sym != SDLK_DELETE  ) && 
				(Event.key.keysym.sym != SDLK_INSERT  ) && 
				(Event.key.keysym.sym < 272 ) &&
				(Event.key.keysym.unicode < 272 ) )	
			{
				//printf("%i %i %i\n", Key, Event.key.keysym.sym, Event.key.keysym.unicode );		
				Client->Engine->Key( this, Key );
			}
			else
			{
				KeyRepeatKey = IK_None;
			}

			break;

		case SDL_KEYUP:
			// Get key code.
			if ( Event.key.keysym.unicode ) {
				Key = (EInputKey) Event.key.keysym.unicode;
			} else {
				Key = (EInputKey) Event.key.keysym.sym;
			}
			
			// Convert to UpperCase/LowerCase
			if ((Key >= ((EInputKey) SDLK_a)) && (Key <= ((EInputKey) SDLK_z))) 
			{
				LowerCase = Key;
				UpperCase = (EInputKey) (Key - 32);
			} 
			else if ((Key >= ((EInputKey) (SDLK_a - 32))) && (Key <= ((EInputKey) (SDLK_z - 32)))) 
			{
				UpperCase = Key;
				LowerCase = (EInputKey) (Key + 32);
			}
			else 
			{
				UpperCase = Key;
				LowerCase = Key;
			}
			
			// Always send upper case letters to the system.
			Key = UpperCase;

			// Check the Keysym map.
			if (KeysymMap[Key] != 0)
				Key = (EInputKey) KeysymMap[Key];
		
			// Release all types of this key.
			if (Key ==(EInputKey) SDLK_BACKSPACE)
				Key = IK_Backspace;
			if (Key == (EInputKey) SDLK_TAB)
				Key = IK_Tab;
			if (Key ==(EInputKey) SDLK_RETURN)
				Key = IK_Enter;

			// Send key to input system.
			CauseInputEvent( Key, IST_Release );

			// Turn off repeating. Needed for
			// chat modes.
			KeyRepeatKey = 0;
			break;
		
		case SDL_MOUSEBUTTONDOWN:
			switch (Event.button.button)
			{
				case 1:
					Key = IK_LeftMouse;
					break;
				case 2:
					Key = IK_MiddleMouse;
					break;
				case 3:
					Key = IK_RightMouse;
					break;
				case 4:
					Key = IK_MouseWheelUp;
					break;
				case 5:
					Key = IK_MouseWheelDown;
					break;
			}
			// Send to input system.
			CauseInputEvent( Key, IST_Press );
			break;

		case SDL_MOUSEBUTTONUP:
			switch (Event.button.button)
			{
				case 1:
					Key = IK_LeftMouse;
					break;
				case 2:
					Key = IK_MiddleMouse;
					break;
				case 3:
					Key = IK_RightMouse;
					break;
				case 4:
					Key = IK_MouseWheelUp;
						break;
				case 5:
					Key = IK_MouseWheelDown;
					break;
			}
			// Send to input system.
			CauseInputEvent( Key, IST_Release );
			break;

		case SDL_MOUSEMOTION:
			MouseMoved = true;		
			DX += Event.motion.xrel;
			DY += Event.motion.yrel;
			break;

		case SDL_JOYBALLMOTION:
			if ( Event.jball.which != Client->JoystickNumber || !Client->UseJoystick )
				break;
			MouseMoved = true;
			DX += (INT) (Event.jball.xrel * Client->ScaleJBX);
			DY += (INT) (Event.jball.yrel * Client->ScaleJBY);
                        break;

		case SDL_JOYBUTTONDOWN:
			if ( Event.jbutton.which != Client->JoystickNumber || !Client->UseJoystick )
				break;
			if ( !JoystickButtonMap[ Event.jbutton.button ] )
			{
				CauseInputEvent( IK_Joy1 + Event.jbutton.button, IST_Press );
				JoystickButtonMap[ Event.jbutton.button ] = 1;
			}
			break;

		case SDL_JOYBUTTONUP:
			if ( Event.jbutton.which != Client->JoystickNumber || !Client->UseJoystick )
				break;
			if ( JoystickButtonMap[ Event.jbutton.button ] )
			{
				CauseInputEvent( IK_Joy1 + Event.jbutton.button, IST_Release );
				JoystickButtonMap[ Event.jbutton.button ] = 0;
			}
			break;

		case SDL_JOYHATMOTION:
			if ( Event.jhat.which != Client->JoystickNumber || !Client->UseJoystick )
				break;
			JoyHatMoved = true;
	
			switch ( Event.jhat.value )
			{
			case SDL_HAT_UP :
				ThisJoyHat = IK_JoyPovUp;		
				break;
			case SDL_HAT_DOWN :
				ThisJoyHat = IK_JoyPovDown;
				break;
			case SDL_HAT_LEFT :
				ThisJoyHat = IK_JoyPovLeft;
				break;
			case SDL_HAT_RIGHT :
				ThisJoyHat = IK_JoyPovRight;
				break;
			default :
				ThisJoyHat = (EInputKey) 0;
				break;
			}				
			break;

		default:;
		}
	}

	if ( Iconified )
	{
		//printf("ICONIFIED\n");
		return;
	}

	// Deliver mouse behavior to the engine.
	if ( MouseMoved )
	{
		// Send to input subsystem.
		if( DX )
			CauseInputEvent( IK_MouseX, IST_Axis, +DX );
		      
		if( DY )
			CauseInputEvent( IK_MouseY, IST_Axis, -DY );

		if (!Client->InMenuLoop)
			Client->Engine->MouseDelta(this, 0, DX, DY);
	}

	if ( Client->UseJoystick )
	{
		INT NumAxes = Client->NumAxes > 6 ? 6 : Client->NumAxes;
		switch ( NumAxes )
		{
		case 6:
			JoystickInputEvent(
				SDL_JoystickGetAxis(Client->Joystick, 5), 
				IK_JoyV, 
				Client->ScaleV, 
				Client->DeadZoneV );
		case 5:
			JoystickInputEvent(
				SDL_JoystickGetAxis(Client->Joystick, 4), 
				IK_JoyU, 
				Client->ScaleU, 
				Client->DeadZoneU );
		
		case 4:
			JoystickInputEvent(
				SDL_JoystickGetAxis(Client->Joystick, 3), 
				IK_JoyR, 
				Client->ScaleR, 
				Client->DeadZoneR );

		case 3:
			JoystickInputEvent(
				SDL_JoystickGetAxis(Client->Joystick, 2), 
				IK_JoyZ, 
				Client->ScaleZ, 
				Client->DeadZoneZ );
		case 2:
			JoystickInputEvent( 
				SDL_JoystickGetAxis(Client->Joystick, 1), 
				IK_JoyY, 
				Client->ScaleY * (Client->InvertVertical ? 1.0 : -1.0), 
				Client->DeadZoneY );

		case 1:
			JoystickInputEvent( 
				SDL_JoystickGetAxis(Client->Joystick, 0), 
				IK_JoyX, 
				Client->ScaleX, 
				Client->DeadZoneX );
		default:
			break;
		}
	}	

	if ( (LastJoyHat != ThisJoyHat) && JoyHatMoved )
	{
		if (LastJoyHat)
			CauseInputEvent( LastJoyHat, IST_Release );
		if (ThisJoyHat)	
			CauseInputEvent( ThisJoyHat, IST_Press );		
		LastJoyHat = ThisJoyHat;
	} 


	// Send WM_CHAR for down keys.
	if ( LastKey && ( appSeconds() - RepeatTimer < 0.5 ) ) 
		return;
	LastKey = false;
	if ( appSeconds() - RepeatTimer < 0.1 )
		return;
		
	RepeatTimer = appSeconds();

	// Deliver keyboard events to the engine	
	if ( KeyRepeatKey )
	{
		if ( KeyRepeatKey == IK_Backspace )
		{
			CauseInputEvent( IK_Backspace, IST_Press );
			CauseInputEvent( IK_Backspace, IST_Release );
		}
		else
 			Client->Engine->Key( this, (EInputKey) KeyRepeatKey );
	}		

	unguard;
}



void USDLViewport::Iconify()
{
	guard(USDLViewport::Iconify);

	if (Iconified)
		return;
		
	Iconified = true;

	// Pause the game if applicable.
        if( GIsRunning )
                Exec( TEXT("SETPAUSE 1"), *this );

        // Release the mouse.
        SetMouseCapture( 0, 0, 0 );
        SetDrag( 0 );

        // Reset the input buffer.
        Input->ResetInput();

        // End fullscreen.
        if( BlitFlags & BLIT_Fullscreen )
                EndFullscreen();
        GetOuterUClient()->MakeCurrent( NULL );

        // Iconify the window.
        SDL_WM_IconifyWindow();

	unguard;
}

/*-----------------------------------------------------------------------------
	Lock and Unlock.
-----------------------------------------------------------------------------*/

//
// Lock the viewport window and set the approprite Screen and RealScreen fields
// of Viewport.  Returns 1 if locked successfully, 0 if failed.  Note that a
// lock failing is not a critical error; it's a sign that a DirectDraw mode
// has ended or the user has closed a viewport window.
//
UBOOL USDLViewport::Lock( FPlane FlashScale, FPlane FlashFog, FPlane ScreenClear, DWORD RenderLockFlags, BYTE* HitData, INT* HitSize )
{
	guard(USDLViewport::LockWindow);
	USDLClient* Client = GetOuterUSDLClient();
	clock(Client->DrawCycles);

	// Success here, so pass to superclass.
	unclock(Client->DrawCycles);
	return UViewport::Lock(FlashScale,FlashFog,ScreenClear,RenderLockFlags,HitData,HitSize);

	unguard;
}

//
// Unlock the viewport window.  If Blit=1, blits the viewport's frame buffer.
//
void USDLViewport::Unlock( UBOOL Blit )
{
	guard(USDLViewport::Unlock);

	UViewport::Unlock( Blit );

	unguard;
}

/*-----------------------------------------------------------------------------
	Viewport modes.
-----------------------------------------------------------------------------*/

//
// Try switching to a new rendering device.
//
void USDLViewport::TryRenderDevice( const TCHAR* ClassName, INT NewX, INT NewY, INT NewColorBytes, UBOOL Fullscreen )
{
	guard(USDLViewport::TryRenderDevice);

	// Shut down current rendering device.
	if( RenDev )
	{
		RenDev->Exit();
		delete RenDev;
		RenDev = NULL;
	}

	// Use appropriate defaults.
	USDLClient* C = GetOuterUSDLClient();
	if( NewX==INDEX_NONE )
		NewX = Fullscreen ? C->FullscreenViewportX : C->WindowedViewportX;
	if( NewY==INDEX_NONE )
		NewY = Fullscreen ? C->FullscreenViewportY : C->WindowedViewportY;

	// Find device driver.
	UClass* RenderClass = UObject::StaticLoadClass( URenderDevice::StaticClass(), NULL, ClassName, NULL, 0, NULL );
	if( RenderClass )
	{
		debugf( TEXT("Loaded render device class.") );
		HoldCount++;
		RenDev = ConstructObject<URenderDevice>( RenderClass, this );
		if( RenDev->Init( this, NewX, NewY, NewColorBytes, Fullscreen ) )
		{
			if( GIsRunning )
				Actor->GetLevel()->DetailChange( RenDev->HighDetailActors );
		}
		else
		{
			debugf( NAME_Log, LocalizeError("Failed3D") );
			delete RenDev;
			RenDev = NULL;
		}
		HoldCount--;
	}
	GRenderDevice = RenDev;
	unguard;
}

//
// If in fullscreen mode, end it and return to Windows.
//
void USDLViewport::EndFullscreen()
{
	guard(USDLViewport::EndFullscreen);
	//USDLClient* Client = GetOuterUSDLClient();
	debugf(NAME_Log, TEXT("Ending fullscreen mode by request."));
	if( RenDev && RenDev->FullscreenOnly )
	{
		// This device doesn't support fullscreen, so use a window-capable rendering device.
		TryRenderDevice( TEXT("ini:Engine.Engine.WindowedRenderDevice"), INDEX_NONE, INDEX_NONE, ColorBytes, 0 );
		check(RenDev);
	}
	else if( RenDev && (BlitFlags & BLIT_OpenGL) )
	{
		RenDev->SetRes( INDEX_NONE, INDEX_NONE, ColorBytes, 0 );
	}
	else
	{
		ResizeViewport( BLIT_DibSection );
	}
	UpdateWindowFrame();
	unguard;
}

//
// Toggle fullscreen.
//
void USDLViewport::ToggleFullscreen()
{
	guard(USDLViewport::ToggleFullscreen);
	if( BlitFlags & BLIT_Fullscreen )
	{
		EndFullscreen();
	}
	else if( !(Actor->ShowFlags & SHOW_ChildWindow) )
	{
		debugf(TEXT("AttemptFullscreen"));
		TryRenderDevice( TEXT("ini:Engine.Engine.GameRenderDevice"), INDEX_NONE, INDEX_NONE, ColorBytes, 1 );
		if( !RenDev )
			TryRenderDevice( TEXT("ini:Engine.Engine.WindowedRenderDevice"), INDEX_NONE, INDEX_NONE, ColorBytes, 1 );
		if( !RenDev )
			TryRenderDevice( TEXT("ini:Engine.Engine.WindowedRenderDevice"), INDEX_NONE, INDEX_NONE, ColorBytes, 0 );
	}
	unguard;
}

//
// Resize the viewport.
//
UBOOL USDLViewport::ResizeViewport( DWORD NewBlitFlags, INT InNewX, INT InNewY, INT InNewColorBytes )
{
	guard(USDLViewport::ResizeViewport);
	USDLClient* Client = GetOuterUSDLClient();

	debugf( TEXT("Resizing SDL viewport. X: %i Y: %i"), InNewX, InNewY );

	// Remember viewport.
	UViewport* SavedViewport = NULL;
	if( Client->Engine->Audio && !GIsEditor && !(GetFlags() & RF_Destroyed) )
		SavedViewport = Client->Engine->Audio->GetViewport();

	// Accept default parameters.
	INT NewX          = InNewX         ==INDEX_NONE ? SizeX      : InNewX;
	INT NewY          = InNewY         ==INDEX_NONE ? SizeY      : InNewY;
	INT NewColorBytes = InNewColorBytes==INDEX_NONE ? ColorBytes : InNewColorBytes;

	INT VideoFlags = 0;
	INT VideoBPP   = 0;

	// Pull current driver string.
	FString CurrentDriver = RenDev->GetClass()->GetPathName();

	// Checking whether to use Glide specific settings or OpenGL.
	if ( !appStrcmp( *CurrentDriver, TEXT("GlideDrv.GlideRenderDevice") ) ) 
	{
		// Don't use DGA in conjunction with Glide unless user overrides it.
		if ( getenv( "SDL_VIDEO_X11_DGAMOUSE" ) == NULL )
			putenv( "SDL_VIDEO_X11_DGAMOUSE=0" );
		NewBlitFlags |= BLIT_Fullscreen;
		debugf( TEXT("GlideDrv.GlideRenderDevice") );	
	} 
	else if ( !appStrcmp( *CurrentDriver, TEXT("SDLSoftDrv.SDLSoftwareRenderDevice") ) )
	{		
		debugf( TEXT("SDLSoftDrv.SDLSoftwareRenderDevice") );
		VideoBPP = 16;
		ColorBytes = 2;
		Caps |= CC_RGB565;

	} else
	{
		VideoFlags |= SDL_OPENGL;
		debugf( TEXT("OpenGL") );
	}

	if( NewBlitFlags & BLIT_Fullscreen )
	{
		VideoFlags |= SDL_FULLSCREEN;
	}

	SetMouseCapture( 1, 1, 0 );

	// Default resolution handling.
	NewX = InNewX!=INDEX_NONE ? InNewX : (NewBlitFlags&BLIT_Fullscreen) ? Client->FullscreenViewportX : Client->WindowedViewportX;
	NewY = InNewX!=INDEX_NONE ? InNewY : (NewBlitFlags&BLIT_Fullscreen) ? Client->FullscreenViewportY : Client->WindowedViewportY;

	if ( SDL_SetVideoMode( NewX, NewY, VideoBPP, VideoFlags ) == NULL ) 
	{
		appErrorf( TEXT("Couldn't set video mode: %s\n"), SDL_GetError( ) );
			appExit( );
	}	

	// Window has been created.
	ViewportStatus = SDL_ViewportNormal;

	// Make this viewport current and update its title bar.
	GetOuterUClient()->MakeCurrent( this );	

	// Update audio.
	if( SavedViewport && SavedViewport!=Client->Engine->Audio->GetViewport() )
		Client->Engine->Audio->SetViewport( SavedViewport );

	// Update the window.
	UpdateWindowFrame();

	// Set new info.
	// DWORD OldBlitFlags = BlitFlags;
	BlitFlags          = NewBlitFlags & ~BLIT_ParameterFlags;
	SizeX              = NewX;
	SizeY              = NewY;
	ColorBytes         = NewColorBytes ? NewColorBytes : ColorBytes;

	// Save info.
	if( RenDev && !GIsEditor )
	{
		if( NewBlitFlags & BLIT_Fullscreen )
		{
			if( NewX && NewY )
			{
				Client->FullscreenViewportX = NewX;
				Client->FullscreenViewportY = NewY;
			}
		}
		else
		{
			if( NewX && NewY )
			{
				Client->WindowedViewportX = NewX;
				Client->WindowedViewportY = NewY;
			}
		}
		Client->SaveConfig();
	}
	return 1;
	unguard;
}

void USDLViewport::Tick()
{
	guard(USDLViewport::Tick);
	unguard;
}

void USDLViewport::CaptureInputs()
{
	guard(USDLViewport::CaptureInputs);
	
	SDL_ShowCursor( false );
	if ( GetOuterUSDLClient()->CaptureMouse )
		SDL_WM_GrabInput( SDL_GRAB_ON );
	
	unguard;
}

void USDLViewport::ReleaseInputs()
{
	guard(USDLViewport::ReleaseInputs); 

	// SDL_ShowCursor( true );
	// SDL_WM_GrabInput( SDL_GRAB_OFF );

	unguard;
}

/*-----------------------------------------------------------------------------
	The End.
-----------------------------------------------------------------------------*/
