/*=============================================================================
	SDLGL.cpp: Unreal OpenGL support implementation for SDL.
	Copyright 1999 Epic Games, Inc. All Rights Reserved.

	Other URenderDevice subclasses include:
	* USoftwareRenderDevice: Software renderer.
	* UGlideRenderDevice: 3dfx Glide renderer.
	* UDirect3DRenderDevice: Direct3D renderer.
	* UOpenGLRenderDevice: OpenGL renderer.

	Revision history:
	* Created by Daniel Vogel (based on XMesaGLDrv).
	* Changed by John Fulmer, Jeroen Janssen
	* Major changes by Daniel Vogel

TODO:
	- implement macro textures

=============================================================================*/

#define FANCY_FISHEYE 0

#include "SDLGLDrv.h"
#include <math.h>

/*-----------------------------------------------------------------------------
	Globals.
-----------------------------------------------------------------------------*/

BYTE ScaleByteNormal  [256];		// modulation
BYTE ScaleByteBrighten[256];		// normal
BYTE ScaleByteGamma   [256];		// gamma
FPlane One4(1,1,1,1);			// 4 consecutive floats for OpenGL
#define OVERBRIGHT 1.4f
	
/*-----------------------------------------------------------------------------
	USDLGLDrv.
-----------------------------------------------------------------------------*/

//
// An OpenGL rendering device attached to a SDL viewport.
//
class USDLGLRenderDevice : public URenderDevice
{
	DECLARE_CLASS(USDLGLRenderDevice,URenderDevice,CLASS_Config,SDLGLDrv)

	// Information about a cached texture.
	struct FCachedTexture
	{
		GLuint Id;
		INT BaseMip;
		INT UBits, VBits;
		INT UCopyBits, VCopyBits;
		FPlane ColorNorm, ColorRenorm;
		INT PolyFlags;			// vogel: to detect changes in PolyFlags (used for gamma)
	};

	// Permanent variables.
	UBOOL WasFullscreen;
	TMap<QWORD,FCachedTexture> LocalBindMap, *BindMap;
	TArray<FPlane> Modes;
	UViewport* Viewport;

	// Timing.
	DWORD BindCycles, ImageCycles, ComplexCycles, GouraudCycles, TileCycles;

	// Hardware constraints.
	INT MaxLogUOverV;
	INT MaxLogVOverU;
	INT MinLogTextureSize;
	INT MaxLogTextureSize;
	UBOOL UseZTrick;
	UBOOL UseMultiTexture;
	UBOOL UsePalette;
	UBOOL Multipass1X;
	UBOOL DoPrecache;
	UBOOL ShareLists;
	UBOOL AlwaysMipmap;
	UBOOL UseTrilinear;
	UBOOL UseVertexSpecular;
	UBOOL UseAlphaPalette;
	UBOOL UseS3TC;
	UBOOL UseModulatedGamma;	
	UBOOL UseGammaExtension;
	UBOOL Use4444Textures;
	UBOOL UseTNT; // REMOVE ME - HACK!

	INT AllocatedTextures;

	// Hit info.
	BYTE* HitData;
	INT* HitSize;

	// Lock variables.
	UBOOL ZTrickToggle;
	INT ZTrickFunc;
	FPlane FlashScale, FlashFog;
	FLOAT RProjZ, Aspect;
	DWORD CurrentPolyFlags;
	TArray<INT> GLHitData;
	struct FTexInfo
	{
		QWORD CurrentCacheID;
		FLOAT UMult;
		FLOAT VMult;
		FLOAT UPan;
		FLOAT VPan;
		FPlane ColorNorm;
		FPlane ColorRenorm;
	} TexInfo[4];
	FLOAT RFX2, RFY2;
	FLOAT FrameX, FrameY;
	
	FLOAT GammaCorrection;

#if FANCY_FISHEYE
#endif

	// Static variables.
	static TMap<QWORD,FCachedTexture> SharedBindMap;
	static INT NumDevices;
	static INT LockCount;
	static UBOOL GLLoaded;

	// GL functions.
	#define GL_EXT(name) static UBOOL SUPPORTS##name;
	#define GL_PROC(ext,ret,func,parms) static ret (STDCALL *func)parms;
	#include "SDLGLFuncs.h"
	#undef GL_EXT
	#undef GL_PROC
	
	// Implementation.
	void GenerateGammaTable( FLOAT GammaCorrection )
	{
		// Init gamma tables.
		for( INT i=0; i<256; i++ )
		{
			FLOAT F              = i/255.0f;		
			ScaleByteNormal  [i] = i;
			ScaleByteBrighten[i] = (BYTE)(Clamp((int) (255.0 * 
						pow( F , 1.0 / (2.50 * (GammaCorrection + (Multipass1X ? 0.5 : 0) )) )) , 0, 255));
			ScaleByteGamma   [i] = (BYTE)(Clamp((int) (255.0 * 
						pow( F , 1.0 / (2.50 * (GammaCorrection + (Multipass1X ? 1.0 : 0.5) )) )) , 0, 255));
		}
		if ( UseGammaExtension )
			SDL_SetGamma( 0.8 + 1.5 * GammaCorrection, 0.8 + 1.5 * GammaCorrection, 0.8 + 1.5 * GammaCorrection );		
	}

	UBOOL FindExt( const TCHAR* Name )
	{
		guard(USDLGLRenderDevice::FindExt);
		UBOOL Result = appStrfind(appFromAnsi((char*)glGetString(GL_EXTENSIONS)),Name)!=NULL;
		if( Result )
			debugf( NAME_Init, TEXT("Device supports: %s"), Name );
		return Result;
		unguard;
	}
	
	void FindProc( void*& ProcAddress, char* Name, char* SupportName, UBOOL& Supports, UBOOL AllowExt )
	{
		guard(USDLGLRenderDevice::FindProc);

		if( !ProcAddress )
			ProcAddress = (void*) SDL_GL_GetProcAddress( Name );

		if( !ProcAddress )
		{
			if( Supports )
				debugf( TEXT("   Missing function '%s' for '%s' support"), appFromAnsi(Name), appFromAnsi(SupportName) );
			Supports = 0;
		}
		unguard;
	}
	
	void FindProcs( UBOOL AllowExt )
	{
		guard(USDLGLRenderDevice::FindProcs);
		
		#define GL_EXT(name) if( AllowExt ) SUPPORTS##name = FindExt( TEXT(#name)+1 );
		#define GL_PROC(ext,ret,func,parms) FindProc( *(void**)&func, #func, #ext, SUPPORTS##ext, AllowExt );
		#include "SDLGLFuncs.h"
		#undef GL_EXT
		#undef GL_PROC
		
		unguard;
	}
	
	UBOOL FailedInitf( const TCHAR* Fmt, ... )
	{
		TCHAR TempStr[4096];
		GET_VARARGS( TempStr, ARRAY_COUNT(TempStr), Fmt );
		debugf( NAME_Init, TempStr );
		Exit();
		return 0;
	}
	
	void MakeCurrent()
	{
		guard(USDLGLRenderDevice::MakeCurrent);		
		unguard;
	}
	
	void Check( const char* Tag )
	{
		GLenum Error = glGetError();
		if( Error!=GL_NO_ERROR )
		{
			const TCHAR* Msg;
			switch( Error )
			{
				case GL_INVALID_ENUM:
					Msg = TEXT("GL_INVALID_ENUM");
					break;
				case GL_INVALID_VALUE:
					Msg = TEXT("GL_INVALID_VALUE");
					break;
				case GL_INVALID_OPERATION:
					Msg = TEXT("GL_INVALID_OPERATION");
					break;
				case GL_STACK_OVERFLOW:
					Msg = TEXT("GL_STACK_OVERFLOW");
					break;
				case GL_STACK_UNDERFLOW:
					Msg = TEXT("GL_STACK_UNDERFLOW");
					break;
				case GL_OUT_OF_MEMORY:
					Msg = TEXT("GL_OUT_OF_MEMORY");
					break;
				default :
					Msg = TEXT("UNKNOWN");
			};
			//appErrorf( TEXT("OpenGL Error: %s (%s)"), Msg, Tag );
			debugf( TEXT("OpenGL Error: %s (%s)"), Msg, Tag );
		}
	}
	
	void SetNoTexture( INT Multi )
	{
		guard(USDLGLRenderDevice::SetNoTexture);
		if( TexInfo[Multi].CurrentCacheID != 0 )
		{
			// Set small white texture.
			clock(BindCycles);
			glBindTexture( GL_TEXTURE_2D, 0 );
			TexInfo[Multi].CurrentCacheID = 0;
			unclock(BindCycles);
		}
		unguard;
	}
	
	void SetTexture( INT Multi, FTextureInfo& Info, DWORD PolyFlags, FLOAT PanBias )
	{
		guard(USDLGLRenderDevice::SetTexture);

		// Set panning.
		FTexInfo& Tex = TexInfo[Multi];
		Tex.UPan      = Info.Pan.X + PanBias*Info.UScale;
		Tex.VPan      = Info.Pan.Y + PanBias*Info.VScale;

		// Find in cache.
		if( Info.CacheID==Tex.CurrentCacheID && !Info.bRealtimeChanged )
			return;

		// Make current.
		clock(BindCycles);
		Tex.CurrentCacheID = Info.CacheID;
		FCachedTexture *Bind=BindMap->Find(Info.CacheID), *ExistingBind=Bind;
		if( !Bind )
		{			
			// Figure out OpenGL-related scaling for the texture.
			Bind            = &BindMap->Set( Info.CacheID, FCachedTexture() );
			glGenTextures( 1, &Bind->Id );
			AllocatedTextures++;
			Bind->PolyFlags = PolyFlags;
			Bind->BaseMip   = Min(0,Info.NumMips-1);
			Bind->UCopyBits = 0;
			Bind->VCopyBits = 0;
			Bind->UBits     = Info.Mips[Bind->BaseMip]->UBits;
			Bind->VBits     = Info.Mips[Bind->BaseMip]->VBits;			
			if( Bind->UBits-Bind->VBits > MaxLogUOverV )
			{
				Bind->VCopyBits += (Bind->UBits-Bind->VBits)-MaxLogUOverV;
				Bind->VBits      = Bind->UBits-MaxLogUOverV;
			}
			if( Bind->VBits-Bind->UBits > MaxLogVOverU )
			{
				Bind->UCopyBits += (Bind->VBits-Bind->UBits)-MaxLogVOverU;
				Bind->UBits      = Bind->VBits-MaxLogVOverU;
			}
			if( Bind->UBits < MinLogTextureSize )
			{
				Bind->UCopyBits += MinLogTextureSize - Bind->UBits;
				Bind->UBits     += MinLogTextureSize - Bind->UBits;
			}
			if( Bind->VBits < MinLogTextureSize )
			{
				Bind->VCopyBits += MinLogTextureSize - Bind->VBits;
				Bind->VBits     += MinLogTextureSize - Bind->VBits;
			}
			if( Bind->UBits > MaxLogTextureSize )
			{			
				Bind->BaseMip += Bind->UBits-MaxLogTextureSize;
				Bind->VBits   -= Bind->UBits-MaxLogTextureSize;
				Bind->UBits    = MaxLogTextureSize;
				if( Bind->VBits < 0 )
				{
					Bind->VCopyBits = -Bind->VBits;
					Bind->VBits     = 0;
				}
			}
			if( Bind->VBits > MaxLogTextureSize )
			{			
				Bind->BaseMip += Bind->VBits-MaxLogTextureSize;
				Bind->UBits   -= Bind->VBits-MaxLogTextureSize;
				Bind->VBits    = MaxLogTextureSize;
				if( Bind->UBits < 0 )
				{
					Bind->UCopyBits = -Bind->UBits;
					Bind->UBits     = 0;
				}
			}
		}
		glBindTexture( GL_TEXTURE_2D, Bind->Id );
		unclock(BindCycles);

		// Account for all the impact on scale normalization.
		Tex.UMult = 1.0 / (Info.UScale * (Info.USize << Bind->UCopyBits));
		Tex.VMult = 1.0 / (Info.VScale * (Info.VSize << Bind->VCopyBits));


		// vogel: Unreal doesn't seem to pass those flags the first time the texture gets uploaded!	
		UBOOL PolyFlagsChanged = (UseModulatedGamma || UseGammaExtension) ? 0 : 
					(!(PolyFlags & (PF_RenderFog|PF_Translucent|PF_Modulated|PF_Highlighted)) && 
					(Bind->PolyFlags & (PF_RenderFog|PF_Translucent|PF_Modulated|PF_Highlighted)));

		// Upload if needed.
		if( !ExistingBind || Info.bRealtimeChanged || PolyFlagsChanged )
		{
			// Cleanup texture flags.
			if ( !PolyFlagsChanged || SupportsLazyTextures )
				Info.Load();
			Info.bRealtimeChanged = 0;

			// Keep Track of PolyFlags
			Bind->PolyFlags = PolyFlags;

			// Set maximum color.
//			Info.CacheMaxColor();
			*Info.MaxColor = FColor(255,255,255,255); // Brandon's color hack.
			Bind->ColorNorm = Info.MaxColor->Plane();
			Bind->ColorNorm.W = 1;

			if( Multipass1X )
			{
				Bind->ColorRenorm.X = Min( Bind->ColorNorm.X * OVERBRIGHT, 1.0f );
				Bind->ColorRenorm.Y = Min( Bind->ColorNorm.Y * OVERBRIGHT, 1.0f );
				Bind->ColorRenorm.Z = Min( Bind->ColorNorm.Z * OVERBRIGHT, 1.0f );
			}
			else Bind->ColorRenorm = Bind->ColorNorm;

			// Setup scaling.
			BYTE *ScaleR, *ScaleG, *ScaleB;
			BYTE *GammaR, *GammaG, *GammaB;
	
			if ( ! (UseModulatedGamma || UseGammaExtension) )
			{
				if( PolyFlags & (PF_RenderFog|PF_Translucent|PF_Modulated|PF_Highlighted) )
				{
					ScaleR = ScaleByteNormal;
					ScaleG = ScaleByteNormal;
					ScaleB = ScaleByteNormal;
					GammaR = ScaleByteNormal;
					GammaG = ScaleByteNormal;
					GammaB = ScaleByteNormal;
				}
				else
				{
					ScaleR = ScaleByteBrighten;
					ScaleG = ScaleByteBrighten;
					ScaleB = ScaleByteBrighten;
					GammaR = ScaleByteGamma;
					GammaG = ScaleByteGamma;
					GammaB = ScaleByteGamma;
				}	
			} 
			else 
			{
				ScaleR = ScaleByteNormal;
				ScaleG = ScaleByteNormal;
				ScaleB = ScaleByteNormal;
				GammaR = ScaleByteNormal;
				GammaG = ScaleByteNormal;
				GammaB = ScaleByteNormal;	
			}

			// Generate the palette.
			FColor LocalPal[256], *NewPal=Info.Palette, TempColor(0,0,0,0);	
			UBOOL Paletted;
			if ( UseAlphaPalette )
				Paletted = UsePalette && Info.Palette;
			else
				Paletted = UsePalette && Info.Palette && !(PolyFlags & PF_Masked) && Info.Palette[0].A==255;

			if( Info.Palette )
			{
				TempColor = Info.Palette[0];
				if( PolyFlags & PF_Masked )
					Info.Palette[0] = FColor(0,0,0,0);
				NewPal = LocalPal;
				for( INT i=0; i<256; i++ )
				{
					FColor& Src = Info.Palette[i];
					NewPal[i].R = ScaleR[Src.R];
					NewPal[i].G = ScaleG[Src.G];
					NewPal[i].B = ScaleB[Src.B];
					NewPal[i].A = Src.A;
				}
				if( Paletted )
					glColorTableEXT( GL_TEXTURE_2D, GL_RGBA, 256, GL_RGBA, GL_UNSIGNED_BYTE, NewPal );
			}

			// Download the texture.
			clock(ImageCycles);
			FMemMark Mark(GMem);
			BYTE* Compose  = New<BYTE>( GMem, (1<<(Bind->UBits+Bind->VBits))*4 );
			UBOOL SkipMipmaps = Info.NumMips==1 && !AlwaysMipmap;
		
			//INT MaxLevel = SupportsTC ? Min(Bind->UBits,Bind->VBits) - MinLogTextureSize : Max(Bind->UBits,Bind->VBits);
			//INT MaxLevel = Min(Bind->UBits,Bind->VBits) - MinLogTextureSize;
			INT MaxLevel;
			if ( UseTNT )
				MaxLevel = Max(Bind->UBits,Bind->VBits);
			else
				MaxLevel = Min(Bind->UBits,Bind->VBits) - MinLogTextureSize;

			for( INT Level=0; Level<=MaxLevel; Level++ )
			{
				// Convert the mipmap.
				INT MipIndex=Bind->BaseMip+Level, StepBits=0;
				if( MipIndex>=Info.NumMips )
				{
					StepBits = MipIndex - (Info.NumMips - 1);
					MipIndex = Info.NumMips - 1;
				}
				FMipmapBase* Mip      = Info.Mips[MipIndex];
				DWORD        Mask     = Mip->USize-1;			
				BYTE* 	     Src      = (BYTE*)Compose;
				GLuint       SourceFormat, InternalFormat;
				if( Mip->DataPtr )
				{					
					if( (Info.Format == TEXF_DXT1) && SupportsTC )
					{
						guard(CompressedTexture_S3TC);
						InternalFormat = GL_COMPRESSED_RGB_S3TC_DXT1_EXT;				       
						unguard;
					}
					else if( Paletted )
					{
						guard(ConvertP8_P8);
						SourceFormat   = GL_COLOR_INDEX;
						InternalFormat = GL_COLOR_INDEX8_EXT;
						BYTE* Ptr      = (BYTE*)Compose;
						for( INT i=0; i<(1<<Max(0,Bind->VBits-Level)); i++ )
						{
							BYTE* Base = (BYTE*)Mip->DataPtr + ((i<<StepBits)&(Mip->VSize-1))*Mip->USize;
							for( INT j=0; j<(1<<Max(0,Bind->UBits-Level+StepBits)); j+=(1<<StepBits) )
								*Ptr++ = Base[j&Mask];
						}
						unguard;
					}
					else if( Info.Palette )
					{
						guard(ConvertP8_RGBA8888);					     
						SourceFormat   = GL_RGBA;
						if ( Use4444Textures )
							InternalFormat = GL_RGBA4;
						else
							InternalFormat = GL_RGBA8; //4; // vogel: customers complain about banding in menues
						FColor* Ptr    = (FColor*)Compose;
						for( INT i=0; i<(1<<Max(0,Bind->VBits-Level)); i++ )
						{
							BYTE* Base = (BYTE*)Mip->DataPtr + ((i<<StepBits)&(Mip->VSize-1))*Mip->USize;
							for( INT j=0; j<(1<<Max(0,Bind->UBits-Level+StepBits)); j+=(1<<StepBits) )
								*Ptr++ = NewPal[Base[j&Mask]];						
						}
						unguard;
					}		
					else
					{
						guard(ConvertBGRA7777_RGBA8888);
						SourceFormat   = GL_RGBA;
						InternalFormat = GL_RGBA8;
						FColor* Ptr    = (FColor*)Compose;
						for( INT i=0; i<(1<<Max(0,Bind->VBits-Level)); i++ )
						{
							FColor* Base = (FColor*)Mip->DataPtr + Min<DWORD>((i<<StepBits)&(Mip->VSize-1),Info.VClamp-1)*Mip->USize;
							for( INT j=0; j<(1<<Max(0,Bind->UBits-Level+StepBits)); j+=(1<<StepBits) )
							{
								FColor& Src = Base[Min<DWORD>(j&Mask,Info.UClamp-1)]; 						
								Ptr->R      = GammaB[ 2 * Src.B];
								Ptr->G      = GammaG[ 2 * Src.G];
								Ptr->B      = GammaR[ 2 * Src.R];
								Ptr->A      = Src.A * 2; // because of 7777 							      
								Ptr++;
							}
						}
						unguard;
					}
				}
				if( ExistingBind )
				{
					if ( (Info.Format == TEXF_DXT1) && SupportsTC )
					{
						guard(glCompressedTexSubImage2D);
						glCompressedTexSubImage2DARB( 
							GL_TEXTURE_2D, 
							Level, 
							0, 
							0, 
							1<<Max(0,Bind->UBits-Level), 
							1<<Max(0,Bind->VBits-Level), 
							InternalFormat,
							(Mip->USize * Mip->VSize)/2,
							Mip->DataPtr );
						unguard;
					}
					else
					{
						guard(glTexSubImage2D);
						glTexSubImage2D( 
							GL_TEXTURE_2D, 
							Level, 
							0, 
							0, 
							1<<Max(0,Bind->UBits-Level), 
							1<<Max(0,Bind->VBits-Level), 
							SourceFormat, 
							GL_UNSIGNED_BYTE, 
							Src );
						unguard;
					}
				}
				else
				{
					if ( (Info.Format == TEXF_DXT1) && SupportsTC )
					{
						guard(glCompressedTexImage2D);					
						glCompressedTexImage2DARB(
							GL_TEXTURE_2D, 
							Level, 
							InternalFormat, 
							1<<Max(0,Bind->UBits-Level), 
							1<<Max(0,Bind->VBits-Level), 
							0, 
							(Mip->USize * Mip->VSize)/2,
							Mip->DataPtr );
						unguard;					
       					}
					else
					{
						guard(glTexImage2D);
						glTexImage2D( 
							GL_TEXTURE_2D, 
							Level, 
							InternalFormat, 
							1<<Max(0,Bind->UBits-Level), 
							1<<Max(0,Bind->VBits-Level), 
							0, 
							SourceFormat, 
							GL_UNSIGNED_BYTE, 
							Src );
						unguard;
					}
				}
				if( SkipMipmaps )
					break;
			}

			Mark.Pop();
			unclock(ImageCycles);

			// Set texture state.
			if( !(PolyFlags & PF_NoSmooth) )
			{
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, SkipMipmaps ? GL_LINEAR : 
					(UseTrilinear ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR_MIPMAP_NEAREST) );
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			}
			else
			{
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, SkipMipmaps ? GL_NEAREST : GL_NEAREST_MIPMAP_NEAREST );
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
			}
		       
			if ( !UseTNT )
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, MaxLevel < 0 ? 0 : MaxLevel);

			// Cleanup.
			if( Info.Palette )
				Info.Palette[0] = TempColor;
			if( SupportsLazyTextures )
				Info.Unload();
		}

		// Copy color norm.
		Tex.ColorNorm   = Bind->ColorNorm;
		Tex.ColorRenorm = Bind->ColorRenorm;

		unguard;
	}
	
	void SetBlend( DWORD PolyFlags )
	{
		guard(USDLGLRenderDevice::SetBlend);

		// Adjust PolyFlags according to Unreal's precedence rules.
		// Allows gouraud-polygonal fog only if specular is supported (1-pass fogging).
		if( (PolyFlags & (PF_RenderFog|PF_Translucent|PF_Modulated))!=PF_RenderFog || !UseVertexSpecular )
			PolyFlags &= ~PF_RenderFog;

		if( !(PolyFlags & (PF_Translucent|PF_Modulated)) )
			PolyFlags |= PF_Occlude;
		else if( PolyFlags & PF_Translucent )
			PolyFlags &= ~PF_Masked;

		// Detect changes in the blending modes.
		DWORD Xor = CurrentPolyFlags^PolyFlags;
		if( Xor & (PF_Translucent|PF_Modulated|PF_Invisible|PF_Occlude|PF_Masked|PF_Highlighted|PF_NoSmooth|PF_RenderFog|PF_Memorized) )
		{
			if( Xor & PF_Masked )
			{
				if ( PolyFlags & PF_Masked )
				{
					glEnable( GL_ALPHA_TEST );
				}
				else
				{
					glDisable( GL_ALPHA_TEST );
				}
			}
			if( Xor&(PF_Invisible|PF_Translucent|PF_Modulated|PF_Highlighted) )
			{
				if( !(PolyFlags & (PF_Invisible|PF_Translucent|PF_Modulated|PF_Highlighted)) )
				{
					glDisable( GL_BLEND );			
				}
				else if( PolyFlags & PF_Invisible )
				{
					glEnable( GL_BLEND );
					glBlendFunc( GL_ZERO, GL_ONE );			
			       	}
				else if( PolyFlags & PF_Translucent )
				{
					glEnable( GL_BLEND );
					glBlendFunc( GL_ONE, GL_ONE_MINUS_SRC_COLOR );				
				}
				else if( PolyFlags & PF_Modulated )
				{
					glEnable( GL_BLEND );
					glBlendFunc( GL_DST_COLOR, GL_SRC_COLOR );			
				}
				else if( PolyFlags & PF_Highlighted )
				{
					glEnable( GL_BLEND );
					glBlendFunc( GL_ONE, GL_ONE_MINUS_SRC_ALPHA );				
				}
			}
			if( Xor & PF_Invisible )
			{
				if ( PolyFlags & PF_Invisible )
					glColorMask( GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE );
				else
					glColorMask( GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE );		
			}
			if( Xor & PF_Occlude )
			{
				if ( PolyFlags & PF_Occlude )
					glDepthMask( GL_TRUE );
				else
					glDepthMask( GL_FALSE );
			}
			if( Xor & PF_NoSmooth )
			{
				//Direct3DDevice7->SetTextureStageState( 0, D3DTSS_MAGFILTER, (PolyFlags & PF_NoSmooth) ? D3DTFG_POINT : D3DTFG_LINEAR );
				//Direct3DDevice7->SetTextureStageState( 0, D3DTSS_MINFILTER, (PolyFlags & PF_NoSmooth) ? D3DTFN_POINT : D3DTFN_LINEAR );
			}
			if( Xor & PF_RenderFog )
			{
				//Direct3DDevice7->SetRenderState( D3DRENDERSTATE_SPECULARENABLE, (PolyFlags&PF_RenderFog)!=0 );
			}
			if( Xor & PF_Memorized )
			{
				// Switch back and forth from multitexturing.
				//Direct3DDevice7->SetTextureStageState( 1, D3DTSS_COLOROP, (PolyFlags&PF_Memorized) ? D3DTOP_MODULATE   : D3DTOP_DISABLE );
				//Direct3DDevice7->SetTextureStageState( 1, D3DTSS_ALPHAOP, (PolyFlags&PF_Memorized) ? D3DTOP_SELECTARG2 : D3DTOP_DISABLE );
			}
			CurrentPolyFlags = PolyFlags;
		}
		unguard;
	}

	// URenderDevice interface.
	UBOOL Init( UViewport* InViewport, INT NewX, INT NewY, INT NewColorBytes, UBOOL Fullscreen )
	{
		guard(USDLGLRenderDevice::Init);

		debugf( TEXT("Initializing SDLGLDrv...") );
		
		// Driver flags.
		SpanBased		= 0;
		FullscreenOnly		= 0;	//Driver dependent.
		SupportsFogMaps		= 1;
		SupportsDistanceFog	= 0;


		// Init global GL.
		if( NumDevices==0 )
		{
			// Clear dynamic load error conditions.
			dlerror();

			// Bind the library.
			FString OpenGLLibName;
			FString Section = TEXT("SDLGLDrv.SDLGLRenderDevice");
			// Default to libGL.so.1 if not defined
			if (!GConfig->GetString( *Section, TEXT("OpenGLLibName"), OpenGLLibName ))
				OpenGLLibName = TEXT("libGL.so.1");

			if ( !GLLoaded )
			{
				// Only call it once as succeeding calls will 'fail'.
				debugf( TEXT("binding %s"), *OpenGLLibName );
				if ( SDL_GL_LoadLibrary( *OpenGLLibName ) == -1 )
					appErrorf( TEXT(SDL_GetError()) );
  				GLLoaded = true;
			}

			SUPPORTS_GL = 1;
			FindProcs( 0 );
			if( !SUPPORTS_GL )
				return 0;		
		}
		NumDevices++;

		BindMap = ShareLists ? &SharedBindMap : &LocalBindMap;
		Viewport = InViewport;
	
		// Try to change resolution to desired.
		if( !SetRes( NewX, NewY, NewColorBytes, Fullscreen ) )
			return FailedInitf( LocalizeError("ResFailed") );

		return 1;
		unguard;
	}
	
	void UnsetRes()
	{
		guard(UOpenGLRenderDevice::UnsetRes);
		Flush( 1 );
		unguard;
	}
	
	UBOOL SetRes( INT NewX, INT NewY, INT NewColorBytes, UBOOL Fullscreen )
	{
		guard(USDLGlRenderDevice::SetRes);
	
		UnsetRes();    

		INT MinDepthBits;
		FString	Section = TEXT("SDLGLDrv.SDLGLRenderDevice");

		// Minimum size of the depth buffer
		if (!GConfig->GetInt( *Section, TEXT("MinDepthBits"), MinDepthBits ))
			MinDepthBits = 16;
		//debugf( TEXT("MinDepthBits = %i"), MinDepthBits );
		// 16 is the bare minimum.
		if ( MinDepthBits < 16 )
			MinDepthBits = 16; 

		INT RequestDoubleBuffer;
		if (!GConfig->GetInt( *Section, TEXT("RequestDoubleBuffer"     ), RequestDoubleBuffer ))
			RequestDoubleBuffer = 1;

		SDL_GL_SetAttribute( SDL_GL_RED_SIZE, NewColorBytes<=2 ? 5 : 8 );
		SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, NewColorBytes<=2 ? 5 : 8 );
		SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, NewColorBytes<=2 ? 5 : 8 );
		SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, MinDepthBits );
		if ( RequestDoubleBuffer )
			SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
		else
			SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 0 );

		// Change window size.
		Viewport->ResizeViewport( Fullscreen ? (BLIT_Fullscreen|BLIT_OpenGL) : (BLIT_HardwarePaint|BLIT_OpenGL), NewX, NewY, NewColorBytes );

		FrameX = NewX;
		FrameY = NewY;

		// Get info and extensions.
		debugf( NAME_Init, TEXT("GL_VENDOR     : %s"), appFromAnsi((ANSICHAR*)glGetString(GL_VENDOR)) );
		debugf( NAME_Init, TEXT("GL_RENDERER   : %s"), appFromAnsi((ANSICHAR*)glGetString(GL_RENDERER)) );
		debugf( NAME_Init, TEXT("GL_VERSION    : %s"), appFromAnsi((ANSICHAR*)glGetString(GL_VERSION)) );
		// vogel: logging of more than 1024 characters is dangerous at the moment. But I still like the info ;-)
		// debugf( NAME_Init, TEXT("GL_EXTENSIONS : %s"), appFromAnsi((ANSICHAR*)glGetString(GL_EXTENSIONS)) );
		printf("GL_EXTENSIONS : %s\n", glGetString(GL_EXTENSIONS) );

		FindProcs( 1 );

		// Set modelview.
		glMatrixMode( GL_MODELVIEW );
		glLoadIdentity();	
		FLOAT Matrix[16] =
		{
			+1, +0, +0, +0,
			+0, -1, +0, +0,
			+0, +0, -1, +0,
			+0, +0, +0, +1,
		};
		glMultMatrixf( Matrix );

		// Get hardware defaults
		GConfig->GetInt( *Section, TEXT("MaxLogUOverV"            ), MaxLogUOverV          );
		GConfig->GetInt( *Section, TEXT("MaxLogVOverU"            ), MaxLogVOverU          );
		GConfig->GetInt( *Section, TEXT("UseZTrick"               ), UseZTrick             );
		GConfig->GetInt( *Section, TEXT("UseMultiTexture"         ), UseMultiTexture       );
		GConfig->GetInt( *Section, TEXT("UsePalette"              ), UsePalette            );
		GConfig->GetInt( *Section, TEXT("ShareLists"              ), ShareLists            );
		GConfig->GetInt( *Section, TEXT("AlwaysMipmap"            ), AlwaysMipmap          );
		GConfig->GetInt( *Section, TEXT("DoPrecache"              ), DoPrecache            );
		GConfig->GetInt( *Section, TEXT("UseTrilinear"            ), UseTrilinear          );			
		GConfig->GetInt( *Section, TEXT("UseAlphaPalette"         ), UseAlphaPalette       );
		GConfig->GetInt( *Section, TEXT("UseS3TC"                 ), UseS3TC               );
		GConfig->GetInt( *Section, TEXT("UseGammaExtension"       ), UseGammaExtension     );
		GConfig->GetInt( *Section, TEXT("UseModulatedGamma"       ), UseModulatedGamma     );
		GConfig->GetInt( *Section, TEXT("Use4444Textures"         ), Use4444Textures       );
		GConfig->GetInt( *Section, TEXT("UseTNT"                  ), UseTNT                );

		if (UseGammaExtension && UseModulatedGamma)
			UseModulatedGamma = 0;

		GammaCorrection = Viewport->GetOuterUClient()->Brightness;
		GenerateGammaTable( GammaCorrection );

		DoPrecache 		= DoPrecache;
		SupportsLazyTextures 	= 0;
		PrefersDeferredLoad	= 0;
		AlwaysMipmap 		= 0; // Doesn't work.
		UseZTrick 		= 0; // Don't even think about halfing the precision of the Z Buffer.
		UseVertexSpecular 	= 0;
		SupportsTC		= UseS3TC;

		// Validate flags.
		if( !SUPPORTS_GL_ARB_multitexture )
			UseMultiTexture = 0;
		if( !SUPPORTS_GL_EXT_paletted_texture )
			UsePalette = 0;
		if( !SUPPORTS_GL_EXT_texture_compression_s3tc && !SUPPORTS_GL_ARB_texture_compression)
		        SupportsTC = 0;

		if ( SupportsTC )
		{
			debugf( TEXT("Trying to use S3TC extension.") );
			if (!(UseGammaExtension || UseModulatedGamma))
				UseGammaExtension = 1;
		}

		// Special treatment for texture size stuff.
		if (!GConfig->GetInt( *Section, TEXT("MinLogTextureSize"), MinLogTextureSize ))
			MinLogTextureSize=0;
	        if (!GConfig->GetInt( *Section, TEXT("MaxLogTextureSize"), MaxLogTextureSize ))
			MaxLogTextureSize=8;

		INT MaxTextureSize;
		glGetIntegerv(GL_MAX_TEXTURE_SIZE, &MaxTextureSize);		
		INT Dummy = -1;
		while (MaxTextureSize > 0)
		{
			MaxTextureSize >>= 1;
			Dummy++;
		}
		
		if ( (MaxLogTextureSize > Dummy) || (UseS3TC) )
			MaxLogTextureSize = Dummy;
		if ( (MinLogTextureSize < 2) || (UseS3TC) )
			MinLogTextureSize = 2;	
		debugf( TEXT("MinLogTextureSize = %i"), MinLogTextureSize );
		debugf( TEXT("MaxLogTextureSize = %i"), MaxLogTextureSize );


		// Verify hardware defaults.
		check(MinLogTextureSize>=0);
		check(MaxLogTextureSize>=0);
		check(MinLogTextureSize<MaxLogTextureSize);
		check(MinLogTextureSize<=MaxLogTextureSize);

		// Misc.
		Multipass1X = UseMultiTexture; //!! and 2X blending isn't supported
		//ScaleByte   = Multipass1X ? ScaleByteBrighten : ScaleByteNormal;

		// Bind little white RGBA texture to ID 0.
		FColor Data[8*8];
		for( INT i=0; i<ARRAY_COUNT(Data); i++ )
			Data[i] = FColor(255,255,255,255);
		SetNoTexture( 0 );
		for( INT Level=0; 8>>Level; Level++ )
			glTexImage2D( GL_TEXTURE_2D, Level, 4, 8>>Level, 8>>Level, 0, GL_RGBA, GL_UNSIGNED_BYTE, Data );

#if FANCY_FISHEYE
#endif		

		// Set permanent state.
		glEnable( GL_DEPTH_TEST );
		glShadeModel( GL_SMOOTH );
		glEnable( GL_TEXTURE_2D );
		glAlphaFunc( GL_GREATER, 0.5 );
		glDepthMask( GL_TRUE );
		glDepthFunc( GL_LEQUAL );
		glBlendFunc( GL_ONE, GL_ZERO );
		glEnable( GL_BLEND );
		glDisable( GL_DITHER );

		CurrentPolyFlags = PF_Occlude;

		// Remember fullscreenness.
		WasFullscreen = Fullscreen;
		return 1;

		unguard;
	}
	
	void Exit()
	{
		guard(USDLGLRenderDevice::Exit);
		check(NumDevices>0);

		// Shut down RC.
		Flush( 1 );
	
		UnsetRes();

		// Shut down global GL.
		if( --NumDevices==0 )
		{
			SharedBindMap.~TMap<QWORD,FCachedTexture>();		
		}

		unguard;
	}
	
	void ShutdownAfterError()
	{
		guard(USDLGLRenderDevice::ShutdownAfterError);
		debugf( NAME_Exit, TEXT("USDLGLRenderDevice::ShutdownAfterError") );
		//ChangeDisplaySettings( NULL, 0 );
		unguard;
	}
	
	void Flush( UBOOL AllowPrecache )
	{
		guard(USDLGLRenderDevice::Flush);
		TArray<GLuint> Binds;
		for( TMap<QWORD,FCachedTexture>::TIterator It(*BindMap); It; ++It )
			Binds.AddItem( It.Value().Id );
		BindMap->Empty();
		if ( Binds.Num() )
			glDeleteTextures( Binds.Num(), (GLuint*)&Binds(0) );
		AllocatedTextures = 0;
		if( AllowPrecache && DoPrecache && !GIsEditor )
			PrecacheOnFlip = 1;

		GammaCorrection = Viewport->GetOuterUClient()->Brightness;
		GenerateGammaTable( GammaCorrection );

		unguard;
	}
	
	static QSORT_RETURN CDECL CompareRes( const FPlane* A, const FPlane* B )
	{
		return (QSORT_RETURN) ( (A->X-B->X)!=0.0 ? (A->X-B->X) : (A->Y-B->Y) );
	}
	
	UBOOL Exec( const TCHAR* Cmd, FOutputDevice& Ar )
	{
		guard(USDLGLRenderDevice::Exec);
		if( ParseCommand(&Cmd,TEXT("GetRes")) )
		{
    			// Changing Resolutions:
			// Entries in the resolution box in the console is
			// apparently controled building a string of relevant 
			// resolutions and sending it to the engine via Ar.Log()

			// Here I am querying SDL_ListModes for available resolutions,
			// and populating the dropbox with its output.
			FString Str = "";
			SDL_Rect **modes;
                        INT i,j;

                	// Available fullscreen video modes
               		modes = SDL_ListModes(NULL, SDL_FULLSCREEN);
                        
			if ( modes == (SDL_Rect **)0 ) 
			{
				debugf( NAME_Init, TEXT("No available fullscreen video modes") );
                	} 
			else if ( modes == (SDL_Rect **)-1 ) 
			{
				debugf( NAME_Init, TEXT("No special fullscreen video modes") );
                	} 
			else 
	    		{
				// count the number of available modes
		    		for ( i=0,j=0; modes[i]; ++i )	
					++j;
						
	    			// Load the string with resolutions from smallest to 
				// largest. SDL_ListModes() provides them from lg
				// to sm...
                    		for ( i=(j-1); i >= 0; --i )
					Str += FString::Printf( TEXT("%ix%i "), modes[i]->w, modes[i]->h);
			}
			
			// Send the resolution string to the engine.	
			Ar.Log( *Str.LeftChop(1) );
			return 1;
		}
		return 0;
		unguard;
	}
	
	void Lock( FPlane InFlashScale, FPlane InFlashFog, FPlane ScreenClear, DWORD RenderLockFlags, BYTE* InHitData, INT* InHitSize )
	{
		guard(USDLGLRenderDevice::Lock);
		check(LockCount==0);

		BindCycles = ImageCycles = ComplexCycles = GouraudCycles = TileCycles = 0;
		++LockCount;

		// Make this context current.
		MakeCurrent();

		// Clear the Z buffer if needed.
		if( !UseZTrick || GIsEditor || (RenderLockFlags & LOCKR_ClearScreen) )
		{
			glClearColor( ScreenClear.X, ScreenClear.Y, ScreenClear.Z, ScreenClear.W );
			glClearDepth( 1.0 );
			glDepthRange( 0.0, 1.0 );
			ZTrickFunc = GL_LEQUAL;
			SetBlend( PF_Occlude );
			glClear( GL_DEPTH_BUFFER_BIT|((RenderLockFlags&LOCKR_ClearScreen)?GL_COLOR_BUFFER_BIT:0) );
		}
		else if( ZTrickToggle )
		{
			ZTrickToggle = 0;
			glClearDepth( 0.5 );
			glDepthRange( 0.0, 0.5 );
			ZTrickFunc = GL_LEQUAL;
		}
		else
		{
			ZTrickToggle = 1;
			glClearDepth( 0.5 );
			glDepthRange( 1.0, 0.5 );
			ZTrickFunc = GL_GEQUAL;
		}
		glDepthFunc( (GLenum) ZTrickFunc );

		// Remember stuff.
		FlashScale = InFlashScale;
		FlashFog   = InFlashFog;
		HitData    = InHitData;
		HitSize    = InHitSize;
		if( HitData )
		{
			*HitSize = 0;
			if( !GLHitData.Num() )
				GLHitData.Add( 16384 );
			glSelectBuffer( GLHitData.Num(), (GLuint*)&GLHitData(0) );
			glRenderMode( GL_SELECT );
			glInitNames();
		}
		unguard;
	}
	
	void SetSceneNode( FSceneNode* Frame )
	{
		guard(USDLGLRenderDevice::SetSceneNode);

		// Precompute stuff.
		Aspect      = Frame->FY/Frame->FX;
		RProjZ      = appTan( Viewport->Actor->FovAngle * PI/360.0 );
		RFX2        = 2.0*RProjZ/Frame->FX;
		RFY2        = 2.0*RProjZ*Aspect/Frame->FY;

		// Set viewport.
		glViewport( Frame->XB, Viewport->SizeY-Frame->Y-Frame->YB, Frame->X, Frame->Y );

		// Set projection.
		glMatrixMode( GL_PROJECTION );
		glLoadIdentity();
		glFrustum( -RProjZ / 2.0, +RProjZ / 2.0, -Aspect*RProjZ / 2.0, +Aspect*RProjZ / 2.0, 1.0 / 2.0, 32768.0 );
	
		// Set clip planes.
		if( HitData )
		{
			FVector N[4];
			N[0] = (FVector((Viewport->HitX-Frame->FX2)*Frame->RProj.Z,0,1)^FVector(0,-1,0)).SafeNormal();
			N[1] = (FVector((Viewport->HitX+Viewport->HitXL-Frame->FX2)*Frame->RProj.Z,0,1)^FVector(0,+1,0)).SafeNormal();
			N[2] = (FVector(0,(Viewport->HitY-Frame->FY2)*Frame->RProj.Z,1)^FVector(+1,0,0)).SafeNormal();
			N[3] = (FVector(0,(Viewport->HitY+Viewport->HitYL-Frame->FY2)*Frame->RProj.Z,1)^FVector(-1,0,0)).SafeNormal();
			for( INT i=0; i<4; i++ )
			{
				double D0[4]={N[i].X,N[i].Y,N[i].Z,0};
				glClipPlane( (GLenum) (GL_CLIP_PLANE0+i), D0 );
				glEnable( (GLenum) (GL_CLIP_PLANE0+i) );
			}
		}
		unguard;
	}

#if FANCY_FISHEYE
#endif
	
	void Unlock( UBOOL Blit )
	{
		guard(USDLGLRenderDevice::Unlock);

		// Unlock and render.
		check(LockCount==1);
		//glFlush();
		if( Blit )
		{
#if FANCY_FISHEYE
#endif
			if ( UseModulatedGamma )
			{
				SetNoTexture( 0 );
				SetBlend( PF_Modulated );			
				glColor4f( GammaCorrection, GammaCorrection, GammaCorrection, 1 );
				glBegin( GL_TRIANGLE_STRIP );
					glTexCoord2f( 0, 0 ); glVertex3f( -1,      -1,      0.5 ); 
					glTexCoord2f( 0, 1 ); glVertex3f( -1,       1,      0.5 );
					glTexCoord2f( 1, 0 ); glVertex3f(  1,      -1,      0.5 );
					glTexCoord2f( 1, 1 ); glVertex3f(  1,       1,      0.5 );
				glEnd();
			}
			Check("NON FATAL: ");
			SDL_GL_SwapBuffers();
		}
		--LockCount;

		// Hits.
		if( HitData )
		{
			INT Records = glRenderMode( GL_RENDER );
			INT* Ptr = &GLHitData(0);
			DWORD BestDepth = MAXDWORD;
			for( INT i=0; i<Records; i++ )
			{
				INT   NameCount = *Ptr++;
				DWORD MinDepth  = *Ptr++;
				DWORD MaxDepth  = *Ptr++;
				if( MinDepth<=BestDepth )
				{
					BestDepth = MinDepth;
					*HitSize = 0;
					for( INT i=0; i<NameCount; )
					{
						INT Count = Ptr[i++];
						for( INT j=0; j<Count; j+=4 )
							*(INT*)(HitData+*HitSize+j) = Ptr[i++];
						*HitSize += Count;
					}
					check(i==NameCount);
				}
				Ptr += NameCount;
				(void)MaxDepth;
			}
			for( i=0; i<4; i++ )
				glDisable( (GLenum) (GL_CLIP_PLANE0+i) );
		}
		unguard;
	}
	
	void DrawComplexSurface( FSceneNode* Frame, FSurfaceInfo& Surface, FSurfaceFacet& Facet )
	{
		guard(USDLGLRenderDevice::DrawComplexSurface);
		check(Surface.Texture);
		clock(ComplexCycles);
		FLOAT UDot = Facet.MapCoords.XAxis | Facet.MapCoords.Origin;
		FLOAT VDot = Facet.MapCoords.YAxis | Facet.MapCoords.Origin;

		// Mutually exclusive effects.
		if( Surface.DetailTexture && Surface.FogMap )
			Surface.DetailTexture = NULL;

		// Modulation chain.
		if( Surface.LightMap && UseMultiTexture )
		{
			// Draw with multitexture.
			SetBlend( Surface.PolyFlags );
			SetTexture( 0, *Surface.LightMap, PF_Modulated, -0.5 );
			glActiveTextureARB( (GLenum) GL_TEXTURE1_ARB );
			SetTexture( 1, *Surface.Texture, Surface.PolyFlags, 0 );
			glEnable( GL_TEXTURE_2D );
			glColor4f( TexInfo[0].ColorRenorm.X*TexInfo[1].ColorRenorm.X, TexInfo[0].ColorRenorm.Y*TexInfo[1].ColorRenorm.Y, TexInfo[0].ColorRenorm.Z*TexInfo[1].ColorRenorm.Z, 1.0 );
			for( FSavedPoly* Poly=Facet.Polys; Poly; Poly=Poly->Next )
			{
				glBegin( GL_TRIANGLE_FAN );
				for( INT i=0; i<Poly->NumPts; i++ )
				{
					FLOAT U = Facet.MapCoords.XAxis | Poly->Pts[i]->Point;
					FLOAT V = Facet.MapCoords.YAxis | Poly->Pts[i]->Point;
					glMultiTexCoord2fARB( (GLenum) GL_TEXTURE0_ARB, (U-UDot-TexInfo[0].UPan)*TexInfo[0].UMult, (V-VDot-TexInfo[0].VPan)*TexInfo[0].VMult );
					glMultiTexCoord2fARB( (GLenum) GL_TEXTURE1_ARB, (U-UDot-TexInfo[1].UPan)*TexInfo[1].UMult, (V-VDot-TexInfo[1].VPan)*TexInfo[1].VMult );
					glVertex3fv( &Poly->Pts[i]->Point.X );
				}
				glEnd();
			}			
			glDisable( GL_TEXTURE_2D );
			glActiveTextureARB( (GLenum) GL_TEXTURE0_ARB );
		}
		else
		{
			// Draw texture.
			SetBlend( Surface.PolyFlags );
			SetTexture( 0, *Surface.Texture, Surface.PolyFlags, 0 );
			FPlane Modulate = TexInfo[0].ColorNorm;
			glColor4fv( Surface.LightMap ? &One4.X : &Modulate.X );
			for( FSavedPoly* Poly=Facet.Polys; Poly; Poly=Poly->Next )
			{
				glBegin( GL_TRIANGLE_FAN );
				for( INT i=0; i<Poly->NumPts; i++ )
				{
					FLOAT U = Facet.MapCoords.XAxis | Poly->Pts[i]->Point;
					FLOAT V = Facet.MapCoords.YAxis | Poly->Pts[i]->Point;
					glTexCoord2f( (U-UDot-TexInfo[0].UPan)*TexInfo[0].UMult, (V-VDot-TexInfo[0].VPan)*TexInfo[0].VMult );
					glVertex3fv( &Poly->Pts[i]->Point.X );
				}
				glEnd();
			}

			// Draw lightmap.
			if( Surface.LightMap )
			{
				SetBlend( PF_Modulated );
				if( Surface.PolyFlags & PF_Masked )
					glDepthFunc( GL_EQUAL );

				SetTexture( 0, *Surface.LightMap, 0, -0.5 );
				Modulate.X *= TexInfo[0].ColorNorm.X;
				Modulate.Y *= TexInfo[0].ColorNorm.Y;
				Modulate.Z *= TexInfo[0].ColorNorm.Z;
				glColor4fv( &Modulate.X );
				for( FSavedPoly* Poly=Facet.Polys; Poly; Poly=Poly->Next )
				{
					glBegin( GL_TRIANGLE_FAN );
					for( INT i=0; i<Poly->NumPts; i++ )
					{
						FLOAT U = Facet.MapCoords.XAxis | Poly->Pts[i]->Point;
						FLOAT V = Facet.MapCoords.YAxis | Poly->Pts[i]->Point;
						glTexCoord2f( (U-UDot-TexInfo[0].UPan)*TexInfo[0].UMult, (V-VDot-TexInfo[0].VPan)*TexInfo[0].VMult );
						glVertex3fv( &Poly->Pts[i]->Point.X );
					}
					glEnd();
				}
				if( Surface.PolyFlags & PF_Masked )
					glDepthFunc( (GLenum) ZTrickFunc );
			}
		}

		// Draw detail texture overlaid, in a separate pass.
		if( Surface.DetailTexture && DetailTextures )
		{
			FLOAT NearZ = 380.0f;
			UBOOL IsDetailing = false;

			for( FSavedPoly* Poly=Facet.Polys; Poly; Poly=Poly->Next )
			{
				UBOOL IsNear[32], CountNear=0;
				for( INT i=0; i<Poly->NumPts; i++ )
				{
					IsNear[i] = Poly->Pts[i]->Point.Z < NearZ ? 1 : 0;
					CountNear += IsNear[i];
				}
				if( CountNear )
				{
					if ( !IsDetailing )
					{
						SetBlend( PF_Modulated );
						SetTexture( 0, *Surface.DetailTexture, PF_Modulated, 0 );
						glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
						IsDetailing = true;
					}
					glBegin( GL_TRIANGLE_FAN );
					for( INT i=0,j=Poly->NumPts-1; i<Poly->NumPts; j=i++ )
					{
						FLOAT U = Facet.MapCoords.XAxis | Poly->Pts[i]->Point;
						FLOAT V = Facet.MapCoords.YAxis | Poly->Pts[i]->Point;
					
						// vogel: this is simplification that works REALLY well
						glColor4f( 1.0, 1.0 , 1.0, (1 - Clamp( Poly->Pts[i]->Point.Z, 0.0f, NearZ ) / NearZ) / 3.3);
					       	glTexCoord2f( (U-UDot-TexInfo[0].UPan)*TexInfo[0].UMult, (V-VDot-TexInfo[0].VPan)*TexInfo[0].VMult );
						glVertex3fv( &Poly->Pts[i]->Point.X );
					}
					glEnd();
				}
			}
			if ( IsDetailing )
			{
				glBlendFunc( GL_DST_COLOR, GL_SRC_COLOR );
			}
		}

		// Draw fog.
		if( Surface.FogMap )
		{
			SetBlend( PF_Highlighted );
			if( Surface.PolyFlags & PF_Masked )
				glDepthFunc( GL_EQUAL );
			SetTexture( 0, *Surface.FogMap, PF_Highlighted, -0.5 );
			glColor4fv( &TexInfo[0].ColorNorm.X );
			for( FSavedPoly* Poly=Facet.Polys; Poly; Poly=Poly->Next )
			{
				glBegin( GL_TRIANGLE_FAN );
				for( INT i=0; i<Poly->NumPts; i++ )
				{
					FLOAT U = Facet.MapCoords.XAxis | Poly->Pts[i]->Point;
					FLOAT V = Facet.MapCoords.YAxis | Poly->Pts[i]->Point;
					glTexCoord2f( (U-UDot-TexInfo[0].UPan)*TexInfo[0].UMult, (V-VDot-TexInfo[0].VPan)*TexInfo[0].VMult );
					glVertex3fv( &Poly->Pts[i]->Point.X );
				}
				glEnd();
			}
			if( Surface.PolyFlags & PF_Masked )
				glDepthFunc( (GLenum) ZTrickFunc );
		}

		// UnrealEd selection.
		if( (Surface.PolyFlags & PF_Selected) && GIsEditor )
		{
			SetNoTexture( 0 );
			SetBlend( PF_Highlighted );
			glBegin( GL_TRIANGLE_FAN );
			glColor4f( 0.0, 0.0, 0.5, 0.5 );
			for( FSavedPoly* Poly=Facet.Polys; Poly; Poly=Poly->Next )
			{
				glBegin( GL_TRIANGLE_FAN );
				for( INT i=0; i<Poly->NumPts; i++ )
					glVertex3fv( &Poly->Pts[i]->Point.X );
				glEnd();
			}
			glEnd();
		}
		unclock(ComplexCycles);
		unguard;
	}
	
	void DrawGouraudPolygon( FSceneNode* Frame, FTextureInfo& Info, FTransTexture** Pts, INT NumPts, DWORD PolyFlags, FSpanBuffer* Span )
	{
		guard(USDLGLRenderDevice::DrawGouraudPolygon);      
		clock(GouraudCycles);

		SetBlend( PolyFlags );
		SetTexture( 0, Info, PolyFlags, 0 );
		glBegin( GL_TRIANGLE_FAN  );
		if( PolyFlags & PF_Modulated )
		{
			glColor4f( TexInfo[0].ColorNorm.X, TexInfo[0].ColorNorm.Y, TexInfo[0].ColorNorm.Z, 1 );
		}
		for( INT i=0; i<NumPts; i++ )
		{
			FTransTexture* P = Pts[i];
			if( !(PolyFlags & PF_Modulated) )
				glColor4f( P->Light.X*TexInfo[0].ColorNorm.X, P->Light.Y*TexInfo[0].ColorNorm.Y, P->Light.Z*TexInfo[0].ColorNorm.Z, 1 );
			glTexCoord2f( P->U*TexInfo[0].UMult, P->V*TexInfo[0].VMult );
			glVertex3fv( &P->Point.X );
		}
		glEnd();

		if( (PolyFlags & (PF_RenderFog|PF_Translucent|PF_Modulated))==PF_RenderFog )
		{			
			SetNoTexture( 0 );
			SetBlend( PF_Highlighted );
			glBegin( GL_TRIANGLE_FAN );
			for( INT i=0; i<NumPts; i++ )
			{
				FTransTexture* P = Pts[i];
				glColor4f( P->Fog.X, P->Fog.Y, P->Fog.Z, P->Fog.W );
				glVertex3fv( &P->Point.X );
			}
			glEnd();	
		}

		unclock(GouraudCycles);
		unguard;
	}
	
	void DrawTile( FSceneNode* Frame, FTextureInfo& Info, FLOAT X, FLOAT Y, FLOAT XL, FLOAT YL, FLOAT U, FLOAT V, FLOAT UL, FLOAT VL, class FSpanBuffer* Span, FLOAT Z, FPlane Color, FPlane Fog, DWORD PolyFlags )
	{
		guard(USDLGLRenderDevice::DrawTile);
		clock(TileCycles);
		if( Info.Palette && Info.Palette[128].A!=255 && !(PolyFlags&PF_Translucent) )
			PolyFlags |= PF_Highlighted;
		SetBlend( PolyFlags );
		SetTexture( 0, Info, PolyFlags, 0 );
		Color.X *= TexInfo[0].ColorNorm.X;
		Color.Y *= TexInfo[0].ColorNorm.Y;
		Color.Z *= TexInfo[0].ColorNorm.Z;
		Color.W  = 1;	
		if ( PolyFlags & PF_Modulated )
			glColor4f( 1.0f, 1.0f, 1.0f, 1.0f );
		else
			glColor4fv( &Color.X );
		glBegin( GL_TRIANGLE_FAN );
		glTexCoord2f( (U   )*TexInfo[0].UMult, (V   )*TexInfo[0].VMult ); glVertex3f( RFX2*Z*(X   -Frame->FX2), RFY2*Z*(Y   -Frame->FY2), Z );
		glTexCoord2f( (U+UL)*TexInfo[0].UMult, (V   )*TexInfo[0].VMult ); glVertex3f( RFX2*Z*(X+XL-Frame->FX2), RFY2*Z*(Y   -Frame->FY2), Z );
		glTexCoord2f( (U+UL)*TexInfo[0].UMult, (V+VL)*TexInfo[0].VMult ); glVertex3f( RFX2*Z*(X+XL-Frame->FX2), RFY2*Z*(Y+YL-Frame->FY2), Z );
		glTexCoord2f( (U   )*TexInfo[0].UMult, (V+VL)*TexInfo[0].VMult ); glVertex3f( RFX2*Z*(X   -Frame->FX2), RFY2*Z*(Y+YL-Frame->FY2), Z );
		glEnd();
		unclock(TileCycles);
		unguard;
	}
	
	void Draw2DLine( FSceneNode* Frame, FPlane Color, DWORD LineFlags, FVector P1, FVector P2 )
	{
		guard(USDLGLRenderDevice::Draw2DLine);
		SetNoTexture( 0 );
		SetBlend( PF_Highlighted );
		glColor3fv( &Color.X );
		glBegin( GL_LINES );
		glVertex3f( RFX2*P1.Z*(P1.X-Frame->FX2), RFY2*P1.Z*(P1.Y-Frame->FY2), P1.Z );
		glVertex3f( RFX2*P2.Z*(P2.X-Frame->FX2), RFY2*P2.Z*(P2.Y-Frame->FY2), P2.Z );
		glEnd();
		unguard;
	}
	
	void Draw3DLine( FSceneNode* Frame, FPlane Color, DWORD LineFlags, FVector P1, FVector P2 )
	{
		guard(USDLGLRenderDevice::Draw3DLine);
		P1 = P1.TransformPointBy( Frame->Coords );
		P2 = P2.TransformPointBy( Frame->Coords );
		if( Frame->Viewport->IsOrtho() )
		{
			// Zoom.
			P1.X = (P1.X) / Frame->Zoom + Frame->FX2;
			P1.Y = (P1.Y) / Frame->Zoom + Frame->FY2;
			P2.X = (P2.X) / Frame->Zoom + Frame->FX2;
			P2.Y = (P2.Y) / Frame->Zoom + Frame->FY2;
			P1.Z = P2.Z = 1;

			// See if points form a line parallel to our line of sight (i.e. line appears as a dot).
			if( Abs(P2.X-P1.X)+Abs(P2.Y-P1.Y)>=0.2 )
				Draw2DLine( Frame, Color, LineFlags, P1, P2 );
			else if( Frame->Viewport->Actor->OrthoZoom < ORTHO_LOW_DETAIL )
				Draw2DPoint( Frame, Color, LINE_None, P1.X-1, P1.Y-1, P1.X+1, P1.Y+1, P1.Z );
		}
		else
		{
			SetNoTexture( 0 );
			SetBlend( PF_Highlighted );
			glColor3fv( &Color.X );
			glBegin( GL_LINES );
			glVertex3fv( &P1.X );
			glVertex3fv( &P2.X );
			glEnd();
		}
		unguard;
	}
	
	void Draw2DPoint( FSceneNode* Frame, FPlane Color, DWORD LineFlags, FLOAT X1, FLOAT Y1, FLOAT X2, FLOAT Y2, FLOAT Z )
	{
		guard(USDLGLRenderDevice::Draw2DPoint);
		SetBlend( PF_Highlighted );
		SetNoTexture( 0 );
		glColor3fv( &Color.X );	// vogel: was 4
		glBegin( GL_TRIANGLE_FAN );
		glVertex3f( RFX2*Z*(X1-Frame->FX2), RFY2*Z*(Y1-Frame->FY2), Z );
		glVertex3f( RFX2*Z*(X2-Frame->FX2), RFY2*Z*(Y1-Frame->FY2), Z );
		glVertex3f( RFX2*Z*(X2-Frame->FX2), RFY2*Z*(Y2-Frame->FY2), Z );
		glVertex3f( RFX2*Z*(X1-Frame->FX2), RFY2*Z*(Y2-Frame->FY2), Z );
		glEnd();
		unguard;
	}
	
	void ClearZ( FSceneNode* Frame )
	{
		guard(USDLGLRenderDevice::ClearZ);
		SetBlend( PF_Occlude );
		glClear( GL_DEPTH_BUFFER_BIT );
		unguard;
	}
	
	void PushHit( const BYTE* Data, INT Count )
	{
		guard(USDLGLRenderDevice::PushHit);
		glPushName( Count );
		for( INT i=0; i<Count; i+=4 )
			glPushName( *(INT*)(Data+i) );
		unguard;
	}
	
	void PopHit( INT Count, UBOOL bForce )
	{
		guard(USDLGLRenderDevice::PopHit);
		glPopName();
		for( INT i=0; i<Count; i+=4 )
			glPopName();
		//!!implement bforce
		unguard;
	}
	
	void GetStats( TCHAR* Result )
	{
		guard(USDLGLRenderDevice::GetStats);
		appSprintf
		(
			Result,
			TEXT("OpenGL stats: Bind=%04.1f Image=%04.1f Complex=%04.1f Gouraud=%04.1f Tile=%04.1f"),
			GSecondsPerCycle*1000 * BindCycles,
			GSecondsPerCycle*1000 * ImageCycles,
			GSecondsPerCycle*1000 * ComplexCycles,
			GSecondsPerCycle*1000 * GouraudCycles,
			GSecondsPerCycle*1000 * TileCycles
		);
		unguard;
	}
	
	void ReadPixels( FColor* Pixels )
	{
		guard(USDLGLRenderDevice::ReadPixels);
		glReadPixels( 0, 0, Viewport->SizeX, Viewport->SizeY, GL_RGBA, GL_UNSIGNED_BYTE, Pixels );
		for( INT i=0; i<Viewport->SizeY/2; i++ )
		{
			for( INT j=0; j<Viewport->SizeX; j++ )
			{
				Exchange( Pixels[j+i*Viewport->SizeX].R, Pixels[j+(Viewport->SizeY-1-i)*Viewport->SizeX].B );
				Exchange( Pixels[j+i*Viewport->SizeX].G, Pixels[j+(Viewport->SizeY-1-i)*Viewport->SizeX].G );
				Exchange( Pixels[j+i*Viewport->SizeX].B, Pixels[j+(Viewport->SizeY-1-i)*Viewport->SizeX].R );
			}
		}
  		unguard;
	}
	
	void EndFlash()
	{
		if( FlashScale!=FPlane(.5,.5,.5,0) || FlashFog!=FPlane(0,0,0,0) )
		{
			SetBlend( PF_Highlighted );
			SetNoTexture( 0 );
			glColor4f( FlashFog.X, FlashFog.Y, FlashFog.Z, 1.0-Min(FlashScale.X*2.f,1.f) );
			FLOAT RFX2 = 2.0*RProjZ       /Viewport->SizeX;
			FLOAT RFY2 = 2.0*RProjZ*Aspect/Viewport->SizeY;
			glBegin( GL_TRIANGLE_FAN );
				glVertex3f( RFX2*(-Viewport->SizeX/2.0), RFY2*(-Viewport->SizeY/2.0), 1.0 );
				glVertex3f( RFX2*(+Viewport->SizeX/2.0), RFY2*(-Viewport->SizeY/2.0), 1.0 );
				glVertex3f( RFX2*(+Viewport->SizeX/2.0), RFY2*(+Viewport->SizeY/2.0), 1.0 );
				glVertex3f( RFX2*(-Viewport->SizeX/2.0), RFY2*(+Viewport->SizeY/2.0), 1.0 );
			glEnd();
		}
	}
	
	void PrecacheTexture( FTextureInfo& Info, DWORD PolyFlags )
	{
		guard(USDLGLRenderDevice::PrecacheTexture);
		SetTexture( 0, Info, PolyFlags, 0 );
		unguard;
	}
};
IMPLEMENT_CLASS(USDLGLRenderDevice);

// Static variables.
INT USDLGLRenderDevice::NumDevices = 0;
INT USDLGLRenderDevice::LockCount  = 0;
TMap<QWORD, USDLGLRenderDevice::FCachedTexture> USDLGLRenderDevice::SharedBindMap;
UBOOL USDLGLRenderDevice::GLLoaded = false;

// OpenGL function pointers.
#define GL_EXT(name) UBOOL USDLGLRenderDevice::SUPPORTS##name=0;
#define GL_PROC(ext,ret,func,parms) ret (STDCALL *USDLGLRenderDevice::func)parms;
#include "SDLGLFuncs.h"
#undef GL_EXT
#undef GL_PROC

/*-----------------------------------------------------------------------------
	The End.
-----------------------------------------------------------------------------*/
